# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2024 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from unittest import TestCase

from sdbus.unittest import IsolatedDbusTestCase
from sdbus.utils.inspect import inspect_dbus_bus, inspect_dbus_path
from sdbus.utils.parse import parse_get_managed_objects

from sdbus import (
    DbusInterfaceCommon,
    DbusInterfaceCommonAsync,
    dbus_property,
    dbus_property_async,
    sd_bus_open_user,
)

TEST_PATH = "/test"


class FooAsync(DbusInterfaceCommonAsync, interface_name="org.foo"):
    @dbus_property_async("x")
    def foo(self) -> int:
        return 1


class BarAsync(DbusInterfaceCommonAsync, interface_name="org.bar"):
    @dbus_property_async("x")
    def bar(self) -> int:
        return 2


class FooBarAsync(FooAsync, BarAsync):
    ...


class Foo(DbusInterfaceCommon, interface_name="org.foo"):
    @dbus_property("x")
    def foo(self) -> int:
        return 1


class Bar(DbusInterfaceCommon, interface_name="org.bar"):
    @dbus_property("x")
    def bar(self) -> int:
        return 2


class FooBar(Foo, Bar):
    ...


MANAGED_OBJECTS_COMBINED = {
    "/test": {
        "org.foo": {"Foo": ("x", 1)},
        "org.bar": {"Bar": ("x", 2)},
    }
}

MANAGED_OBJECTS_SPLIT = {
    "/foo": {
        "org.foo": {"Foo": ("x", 1)},
    },
    "/bar": {
        "org.bar": {"Bar": ("x", 2)},
    },
}

MANAGED_OBJECTS_BOTH = {**MANAGED_OBJECTS_COMBINED, **MANAGED_OBJECTS_SPLIT}


class TestSdbusUtilsParse(TestCase):
    def test_parse_get_managed_objects_async_combined(self) -> None:
        parsed_managed = parse_get_managed_objects(
            FooBarAsync,
            MANAGED_OBJECTS_COMBINED,
        )

        self.assertEqual(1, len(parsed_managed))

        class_type, properties_data = parsed_managed["/test"]
        self.assertEqual(FooBarAsync, class_type)
        self.assertEqual(properties_data["foo"], 1)
        self.assertEqual(properties_data["bar"], 2)

    def test_parse_get_managed_objects_block_combined(self) -> None:
        parsed_managed = parse_get_managed_objects(
            FooBar,
            MANAGED_OBJECTS_COMBINED,
        )

        self.assertEqual(1, len(parsed_managed))

        class_type, properties_data = parsed_managed["/test"]
        self.assertEqual(FooBar, class_type)
        self.assertEqual(properties_data["foo"], 1)
        self.assertEqual(properties_data["bar"], 2)

    def test_parse_get_managed_objects_async_split(self) -> None:
        parsed_managed = parse_get_managed_objects(
            [FooAsync, BarAsync],
            MANAGED_OBJECTS_SPLIT,
        )

        self.assertEqual(2, len(parsed_managed))

        class_type, properties_data = parsed_managed["/foo"]
        self.assertEqual(FooAsync, class_type)
        self.assertEqual(properties_data["foo"], 1)

        class_type, properties_data = parsed_managed["/bar"]
        self.assertEqual(BarAsync, class_type)
        self.assertEqual(properties_data["bar"], 2)

    def test_parse_get_managed_objects_block_split(self) -> None:
        parsed_managed = parse_get_managed_objects(
            [Foo, Bar],
            MANAGED_OBJECTS_SPLIT,
        )

        self.assertEqual(2, len(parsed_managed))

        class_type, properties_data = parsed_managed["/foo"]
        self.assertEqual(Foo, class_type)
        self.assertEqual(properties_data["foo"], 1)

        class_type, properties_data = parsed_managed["/bar"]
        self.assertEqual(Bar, class_type)
        self.assertEqual(properties_data["bar"], 2)

    def test_parse_get_managed_objects_unknown_interface_error(self) -> None:
        with self.assertRaisesRegex(KeyError, "org.foo"):
            parse_get_managed_objects(
                Bar,
                MANAGED_OBJECTS_SPLIT,
            )

    def test_parse_get_managed_objects_unknown_interface_none_reuse(
        self,
    ) -> None:
        parsed_managed = parse_get_managed_objects(
            {BarAsync},
            MANAGED_OBJECTS_SPLIT,
            on_unknown_interface="none",
            on_unknown_member="reuse",
        )

        class_type, properties_data = parsed_managed["/foo"]
        self.assertIsNone(class_type)
        self.assertEqual(properties_data["Foo"], 1)

        class_type, properties_data = parsed_managed["/bar"]
        self.assertEqual(BarAsync, class_type)
        self.assertEqual(properties_data["bar"], 2)

    def test_parse_get_managed_objects_unknown_member_skip(self) -> None:
        parsed_managed = parse_get_managed_objects(
            [Foo],
            MANAGED_OBJECTS_SPLIT,
            on_unknown_interface="none",
            on_unknown_member="ignore",
        )

        self.assertEqual(2, len(parsed_managed))

        class_type, properties_data = parsed_managed["/foo"]
        self.assertEqual(Foo, class_type)
        self.assertEqual(properties_data["foo"], 1)

        class_type, properties_data = parsed_managed["/bar"]
        self.assertIsNone(class_type)
        self.assertEqual(0, len(properties_data))

    def test_parse_get_managed_objects_interface_subset_single(self) -> None:
        parsed_managed = parse_get_managed_objects(
            [FooAsync],
            MANAGED_OBJECTS_COMBINED,
            on_unknown_interface="error",
            on_unknown_member="reuse",
            use_interface_subsets=True,
        )
        class_type, properties_data = parsed_managed["/test"]
        self.assertIs(class_type, FooAsync)
        self.assertEqual(properties_data["foo"], 1)
        self.assertEqual(properties_data["Bar"], 2)

    def test_parse_get_managed_objects_interface_subset_multiple(self) -> None:
        parsed_managed = parse_get_managed_objects(
            [FooAsync, FooBarAsync],
            # FooBarAsync should be prioritized then both interfaces
            # are available on the path.
            MANAGED_OBJECTS_BOTH,
            on_unknown_interface="none",
            on_unknown_member="reuse",
            use_interface_subsets=True,
        )

        class_type, _ = parsed_managed["/test"]
        self.assertIs(class_type, FooBarAsync)
        class_type, _ = parsed_managed["/foo"]
        self.assertIs(class_type, FooAsync)
        class_type, _ = parsed_managed["/bar"]
        self.assertIsNone(class_type)


class TestSdbusUtilsInspect(IsolatedDbusTestCase):
    def test_inspect_dbus_path_block(self) -> None:
        proxy = DbusInterfaceCommon("example.org", TEST_PATH)

        self.assertEqual(inspect_dbus_path(proxy), TEST_PATH)

        new_bus = sd_bus_open_user()

        with self.assertRaisesRegex(LookupError, "is not attached to bus"):
            inspect_dbus_path(proxy, new_bus)

    def test_inspect_dbus_path_async_proxy(self) -> None:
        proxy = DbusInterfaceCommonAsync.new_proxy("example.org", TEST_PATH)

        self.assertEqual(inspect_dbus_path(proxy), TEST_PATH)

        new_bus = sd_bus_open_user()

        with self.assertRaisesRegex(LookupError, "is not attached to bus"):
            inspect_dbus_path(proxy, new_bus)

    def test_inspect_dbus_path_async_local(self) -> None:
        local_obj = FooBarAsync()

        with self.assertRaisesRegex(
            LookupError, "is not exported to any D-Bus",
        ):
            inspect_dbus_path(local_obj)

        local_obj.export_to_dbus(TEST_PATH)

        self.assertEqual(inspect_dbus_path(local_obj), TEST_PATH)

        new_bus = sd_bus_open_user()

        with self.assertRaisesRegex(LookupError, "is not attached to bus"):
            inspect_dbus_path(local_obj, new_bus)

    def test_inspect_attached_bus(self) -> None:
        proxy = DbusInterfaceCommon("example.org", TEST_PATH)

        self.assertIs(inspect_dbus_bus(proxy), self.bus)

        with self.assertRaises(TypeError):
            inspect_dbus_bus(object())  # type: ignore[arg-type]

    def test_inspect_attached_bus_async(self) -> None:
        proxy = DbusInterfaceCommonAsync.new_proxy("example.org", TEST_PATH)

        self.assertIs(inspect_dbus_bus(proxy), self.bus)

        local_obj = FooBarAsync()

        self.assertIsNone(inspect_dbus_bus(local_obj))

        local_obj.export_to_dbus("/")

        self.assertIs(inspect_dbus_bus(local_obj), self.bus)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from inspect import iscoroutinefunction
from types import FunctionType
from typing import TYPE_CHECKING, TypeVar, cast

from .dbus_common_elements import (
    DbusBoundSync,
    DbusMemberSync,
    DbusMethodCommon,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any, Optional

    from .dbus_proxy_sync_interface_base import DbusInterfaceBase

T = TypeVar('T')


class DbusMethodSync(DbusMethodCommon, DbusMemberSync):
    def __get__(self,
                obj: DbusInterfaceBase,
                obj_class: Optional[type[DbusInterfaceBase]] = None,
                ) -> Callable[..., Any]:
        return DbusLocalMethodSync(self, obj)


class DbusLocalMethodSync(DbusBoundSync):
    def __init__(self,
                 dbus_method: DbusMethodSync,
                 interface: DbusInterfaceBase):
        self.dbus_method = dbus_method
        self.interface = interface

        self.__doc__ = dbus_method.__doc__

    def _call_dbus_sync(self, *args: Any) -> Any:
        new_call_message = (
            self.interface._dbus.attached_bus.new_method_call_message(
                self.interface._dbus.service_name,
                self.interface._dbus.object_path,
                self.dbus_method.interface_name,
                self.dbus_method.method_name,
            )
        )
        if args:
            new_call_message.append_data(
                self.dbus_method.input_signature, *args)

        reply_message = self.interface._dbus.attached_bus.call(
            new_call_message)
        return reply_message.get_contents()

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if len(args) == self.dbus_method.num_of_args:
            assert not kwargs, (
                "Passed more arguments than method supports"
                f"Extra args: {kwargs}")
            rebuilt_args: Sequence[Any] = args
        else:
            rebuilt_args = self.dbus_method._rebuild_args(
                self.dbus_method.original_method,
                *args,
                **kwargs)

        return self._call_dbus_sync(*rebuilt_args)


def dbus_method(
    input_signature: str = "",
    result_signature: str = "",
    flags: int = 0,
    method_name: Optional[str] = None,
) -> Callable[[T], T]:
    assert not isinstance(input_signature, FunctionType), (
        "Passed function to decorator directly. "
        "Did you forget () round brackets?"
    )

    def dbus_method_decorator(original_method: T) -> T:
        assert isinstance(original_method, FunctionType)
        assert not iscoroutinefunction(original_method), (
            "Expected NON coroutine function. ",
            "Maybe you wanted to remove 'async' keyword?",
        )
        new_wrapper = DbusMethodSync(
            original_method=original_method,
            method_name=method_name,
            input_signature=input_signature,
            result_signature=result_signature,
            result_args_names=(),
            input_args_names=(),
            flags=flags,
        )

        return cast(T, new_wrapper)

    return dbus_method_decorator



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from asyncio import Queue
from collections.abc import AsyncIterable, AsyncIterator
from contextlib import closing
from types import FunctionType
from typing import TYPE_CHECKING, Generic, TypeVar, cast, overload
from weakref import WeakSet

from .dbus_common_elements import (
    DbusBoundAsync,
    DbusLocalObjectMeta,
    DbusMemberAsync,
    DbusRemoteObjectMeta,
    DbusSignalCommon,
)
from .default_bus import get_default_bus

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any, Optional, Union

    from .dbus_proxy_async_interface_base import DbusInterfaceBaseAsync
    from .sd_bus_internals import SdBus, SdBusMessage, SdBusSlot


T = TypeVar('T')


class DbusSignalAsync(DbusMemberAsync, DbusSignalCommon, Generic[T]):

    def __init__(
        self,
        signal_name: Optional[str],
        signal_signature: str,
        args_names: Sequence[str],
        flags: int,
        original_method: FunctionType
    ):
        super().__init__(
            signal_name,
            signal_signature,
            args_names,
            flags,
            original_method,
        )

        self.local_callbacks: WeakSet[Callable[[T], Any]] = WeakSet()

    @overload
    def __get__(
        self,
        obj: None,
        obj_class: type[DbusInterfaceBaseAsync],
    ) -> DbusSignalAsync[T]:
        ...

    @overload
    def __get__(
        self,
        obj: DbusInterfaceBaseAsync,
        obj_class: type[DbusInterfaceBaseAsync],
    ) -> DbusBoundSignalAsyncBase[T]:
        ...

    def __get__(
        self,
        obj: Optional[DbusInterfaceBaseAsync],
        obj_class: Optional[type[DbusInterfaceBaseAsync]] = None,
    ) -> Union[DbusBoundSignalAsyncBase[T], DbusSignalAsync[T]]:
        if obj is not None:
            dbus_meta = obj._dbus
            if isinstance(dbus_meta, DbusRemoteObjectMeta):
                return DbusProxySignalAsync(self, dbus_meta)
            else:
                return DbusLocalSignalAsync(self, dbus_meta)
        else:
            return self

    async def catch_anywhere(
        self,
        service_name: str,
        bus: Optional[SdBus] = None,
    ) -> AsyncIterable[tuple[str, T]]:
        if bus is None:
            bus = get_default_bus()

        message_queue: Queue[SdBusMessage] = Queue()

        match_slot = await bus.match_signal_async(
            service_name,
            None,
            self.interface_name,
            self.signal_name,
            message_queue.put_nowait,
        )

        with closing(match_slot):
            while True:
                next_signal_message = await message_queue.get()
                signal_path = next_signal_message.path
                assert signal_path is not None
                yield (
                    signal_path,
                    cast(T, next_signal_message.get_contents())
                )


class DbusBoundSignalAsyncBase(DbusBoundAsync, AsyncIterable[T], Generic[T]):
    async def catch(self) -> AsyncIterator[T]:
        raise NotImplementedError
        yield cast(T, None)

    __aiter__ = catch

    async def catch_anywhere(
            self,
            service_name: Optional[str] = None,
            bus: Optional[SdBus] = None,
    ) -> AsyncIterable[tuple[str, T]]:
        raise NotImplementedError
        yield "", cast(T, None)

    def emit(self, args: T) -> None:
        raise NotImplementedError


class DbusProxySignalAsync(DbusBoundSignalAsyncBase[T]):
    def __init__(
        self,
        dbus_signal: DbusSignalAsync[T],
        proxy_meta: DbusRemoteObjectMeta,
    ):
        self.dbus_signal = dbus_signal
        self.proxy_meta = proxy_meta

        self.__doc__ = dbus_signal.__doc__

    async def _register_match_slot(
        self,
        bus: SdBus,
        callback: Callable[[SdBusMessage], Any],
    ) -> SdBusSlot:
        return await bus.match_signal_async(
            self.proxy_meta.service_name,
            self.proxy_meta.object_path,
            self.dbus_signal.interface_name,
            self.dbus_signal.signal_name,
            callback,
        )

    async def catch(self) -> AsyncIterator[T]:
        message_queue: Queue[SdBusMessage] = Queue()

        match_slot = await self._register_match_slot(
            self.proxy_meta.attached_bus,
            message_queue.put_nowait,
        )

        with closing(match_slot):
            while True:
                next_signal_message = await message_queue.get()
                yield cast(T, next_signal_message.get_contents())

    __aiter__ = catch

    async def catch_anywhere(
            self,
            service_name: Optional[str] = None,
            bus: Optional[SdBus] = None,
    ) -> AsyncIterable[tuple[str, T]]:
        if bus is None:
            bus = self.proxy_meta.attached_bus

        if service_name is None:
            service_name = self.proxy_meta.service_name

        message_queue: Queue[SdBusMessage] = Queue()

        match_slot = await bus.match_signal_async(
            service_name,
            None,
            self.dbus_signal.interface_name,
            self.dbus_signal.signal_name,
            message_queue.put_nowait,
        )

        with closing(match_slot):
            while True:
                next_signal_message = await message_queue.get()
                signal_path = next_signal_message.path
                assert signal_path is not None
                yield (
                    signal_path,
                    cast(T, next_signal_message.get_contents())
                )

    def emit(self, args: T) -> None:
        raise RuntimeError("Cannot emit signal from D-Bus proxy.")


class DbusLocalSignalAsync(DbusBoundSignalAsyncBase[T]):
    def __init__(
        self,
        dbus_signal: DbusSignalAsync[T],
        local_meta: DbusLocalObjectMeta,
    ):
        self.dbus_signal = dbus_signal
        self.local_meta = local_meta

        self.__doc__ = dbus_signal.__doc__

    async def catch(self) -> AsyncIterator[T]:
        new_queue: Queue[T] = Queue()

        signal_callbacks = self.dbus_signal.local_callbacks
        try:
            put_method = new_queue.put_nowait
            signal_callbacks.add(put_method)
            while True:
                next_data = await new_queue.get()
                yield next_data
        finally:
            signal_callbacks.remove(put_method)

    __aiter__ = catch

    async def catch_anywhere(
        self,
        service_name: Optional[str] = None,
        bus: Optional[SdBus] = None,
    ) -> AsyncIterable[tuple[str, T]]:
        raise NotImplementedError("TODO")
        yield

    def _emit_dbus_signal(self, args: T) -> None:
        attached_bus = self.local_meta.attached_bus
        if attached_bus is None:
            return

        serving_object_path = self.local_meta.serving_object_path
        if serving_object_path is None:
            return

        signal_message = attached_bus.new_signal_message(
            serving_object_path,
            self.dbus_signal.interface_name,
            self.dbus_signal.signal_name,
        )

        if ((not self.dbus_signal.signal_signature.startswith('('))
            and
                isinstance(args, tuple)):
            signal_message.append_data(
                self.dbus_signal.signal_signature, *args)
        elif self.dbus_signal.signal_signature == '' and args is None:
            ...
        else:
            signal_message.append_data(
                self.dbus_signal.signal_signature, args)

        signal_message.send()

    def emit(self, args: T) -> None:
        self._emit_dbus_signal(args)

        for callback in self.dbus_signal.local_callbacks:
            callback(args)


def dbus_signal_async(
        signal_signature: str = '',
        signal_args_names: Sequence[str] = (),
        flags: int = 0,
        signal_name: Optional[str] = None,
) -> Callable[
    [Callable[[Any], T]],
    DbusSignalAsync[T]
]:
    assert not isinstance(signal_signature, FunctionType), (
        "Passed function to decorator directly. "
        "Did you forget () round brackets?"
    )

    def signal_decorator(
            pseudo_function: Callable[[Any], T]) -> DbusSignalAsync[T]:

        assert isinstance(pseudo_function, FunctionType)
        return DbusSignalAsync(
            signal_name,
            signal_signature,
            signal_args_names,
            flags,
            pseudo_function,
        )

    return signal_decorator



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from collections.abc import Awaitable
from inspect import iscoroutinefunction
from types import FunctionType
from typing import TYPE_CHECKING, Generic, TypeVar, cast, overload
from weakref import ref as weak_ref

from .dbus_common_elements import (
    DbusBoundAsync,
    DbusMemberAsync,
    DbusPropertyCommon,
    DbusPropertyOverride,
    DbusRemoteObjectMeta,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Generator
    from typing import Any, Optional, Union

    from .dbus_proxy_async_interface_base import DbusInterfaceBaseAsync
    from .sd_bus_internals import SdBusMessage


T = TypeVar('T')


class DbusPropertyAsync(DbusMemberAsync, DbusPropertyCommon, Generic[T]):
    def __init__(
            self,
            property_name: Optional[str],
            property_signature: str,
            property_getter: Callable[[DbusInterfaceBaseAsync],
                                      T],
            property_setter: Optional[
                Callable[[DbusInterfaceBaseAsync, T],
                         None]],
            flags: int,

    ) -> None:
        assert isinstance(property_getter, FunctionType)
        super().__init__(
            property_name,
            property_signature,
            flags,
            property_getter,
        )
        self.property_getter: Callable[
            [DbusInterfaceBaseAsync], T] = property_getter
        self.property_setter: Optional[
            Callable[[DbusInterfaceBaseAsync, T],
                     None]] = property_setter
        self.property_setter_is_public: bool = True

        self.__doc__ = property_getter.__doc__

    @overload
    def __get__(
        self,
        obj: None,
        obj_class: type[DbusInterfaceBaseAsync],
    ) -> DbusPropertyAsync[T]:
        ...

    @overload
    def __get__(
        self,
        obj: DbusInterfaceBaseAsync,
        obj_class: type[DbusInterfaceBaseAsync],
    ) -> DbusBoundPropertyAsyncBase[T]:
        ...

    def __get__(
        self,
        obj: Optional[DbusInterfaceBaseAsync],
        obj_class: Optional[type[DbusInterfaceBaseAsync]] = None,
    ) -> Union[DbusBoundPropertyAsyncBase[T], DbusPropertyAsync[T]]:
        if obj is not None:
            dbus_meta = obj._dbus
            if isinstance(dbus_meta, DbusRemoteObjectMeta):
                return DbusProxyPropertyAsync(self, dbus_meta)
            else:
                return DbusLocalPropertyAsync(self, obj)
        else:
            return self

    def setter(self,
               new_set_function: Callable[
                   [Any, T],
                   None],
               ) -> None:
        assert self.property_setter is None, "Setter already defined"
        assert not iscoroutinefunction(new_set_function), (
            "Property setter can't be coroutine",
        )
        self.property_setter = new_set_function

    def setter_private(
        self,
        new_set_function: Callable[
            [Any, T],
            None],
    ) -> None:
        assert self.property_setter is None, "Setter already defined"
        assert not iscoroutinefunction(new_set_function), (
            "Property setter can't be coroutine",
        )
        self.property_setter = new_set_function
        self.property_setter_is_public = False


class DbusBoundPropertyAsyncBase(DbusBoundAsync, Awaitable[T]):
    def __await__(self) -> Generator[Any, None, T]:
        return self.get_async().__await__()

    async def get_async(self) -> T:
        raise NotImplementedError

    async def set_async(self, complete_object: T) -> None:
        raise NotImplementedError


class DbusProxyPropertyAsync(DbusBoundPropertyAsyncBase[T]):
    def __init__(
        self,
        dbus_property: DbusPropertyAsync[T],
        proxy_meta: DbusRemoteObjectMeta,
    ):
        self.dbus_property = dbus_property
        self.proxy_meta = proxy_meta

        self.__doc__ = dbus_property.__doc__

    async def get_async(self) -> T:
        bus = self.proxy_meta.attached_bus
        new_get_message = (
            bus.new_property_get_message(
                self.proxy_meta.service_name,
                self.proxy_meta.object_path,
                self.dbus_property.interface_name,
                self.dbus_property.property_name,
            )
        )
        reply_message = await bus.call_async(new_get_message)
        # Get method returns variant but we only need contents of variant
        return cast(T, reply_message.get_contents()[1])

    async def set_async(self, complete_object: T) -> None:
        bus = self.proxy_meta.attached_bus
        new_set_message = (
            bus.new_property_set_message(
                self.proxy_meta.service_name,
                self.proxy_meta.object_path,
                self.dbus_property.interface_name,
                self.dbus_property.property_name,
            )
        )
        new_set_message.append_data(
            'v',
            (self.dbus_property.property_signature, complete_object),
        )
        await bus.call_async(new_set_message)


class DbusLocalPropertyAsync(DbusBoundPropertyAsyncBase[T]):
    def __init__(
        self,
        dbus_property: DbusPropertyAsync[T],
        local_object: DbusInterfaceBaseAsync,
    ):
        self.dbus_property = dbus_property
        self.local_object_ref = weak_ref(local_object)

        self.__doc__ = dbus_property.__doc__

    async def get_async(self) -> T:
        local_object = self.local_object_ref()
        if local_object is None:
            raise RuntimeError("Local object no longer exists!")

        return self.dbus_property.property_getter(local_object)

    async def set_async(self, complete_object: T) -> None:
        if self.dbus_property.property_setter is None:
            raise RuntimeError("Property has no setter")

        local_object = self.local_object_ref()
        if local_object is None:
            raise RuntimeError("Local object no longer exists!")

        self.dbus_property.property_setter(
            local_object,
            complete_object,
        )

        try:
            properties_changed = getattr(
                local_object,
                "properties_changed",
            )
        except AttributeError:
            ...
        else:
            properties_changed.emit(
                (
                    self.dbus_property.interface_name,
                    {
                        self.dbus_property.property_name: (
                            self.dbus_property.property_signature,
                            complete_object,
                        ),
                    },
                    []
                )
            )

    def _dbus_reply_get(self, message: SdBusMessage) -> None:
        local_object = self.local_object_ref()
        if local_object is None:
            raise RuntimeError("Local object no longer exists!")

        reply_data: Any = self.dbus_property.property_getter(local_object)
        message.append_data(self.dbus_property.property_signature, reply_data)

    def _dbus_reply_set(self, message: SdBusMessage) -> None:
        local_object = self.local_object_ref()
        if local_object is None:
            raise RuntimeError("Local object no longer exists!")

        assert self.dbus_property.property_setter is not None
        data_to_set_to: Any = message.get_contents()

        self.dbus_property.property_setter(local_object, data_to_set_to)

        try:
            properties_changed = getattr(
                local_object,
                "properties_changed",
            )
        except AttributeError:
            ...
        else:
            properties_changed.emit(
                (
                    self.dbus_property.interface_name,
                    {
                        self.dbus_property.property_name: (
                            self.dbus_property.property_signature,
                            data_to_set_to,
                        ),
                    },
                    []
                )
            )


def dbus_property_async(
        property_signature: str = "",
        flags: int = 0,
        property_name: Optional[str] = None,
) -> Callable[
    [Callable[[Any], T]],
        DbusPropertyAsync[T]]:

    assert not isinstance(property_signature, FunctionType), (
        "Passed function to decorator directly. "
        "Did you forget () round brackets?"
    )

    def property_decorator(
        function: Callable[..., Any]
    ) -> DbusPropertyAsync[T]:

        assert not iscoroutinefunction(function), (
            "Property getter can't be coroutine",
        )

        new_wrapper: DbusPropertyAsync[T] = DbusPropertyAsync(
            property_name,
            property_signature,
            function,
            None,
            flags,
        )

        return new_wrapper

    return property_decorator


def dbus_property_async_override() -> Callable[
    [Callable[[Any], T]],
        DbusPropertyAsync[T]]:

    def new_decorator(
            new_property: Callable[[Any], T]) -> DbusPropertyAsync[T]:
        return cast(DbusPropertyAsync[T], DbusPropertyOverride(new_property))

    return new_decorator



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

import threading
from contextvars import ContextVar, Token
from logging import getLogger
from typing import TYPE_CHECKING

from .sd_bus_internals import (
    NameAllowReplacementFlag,
    NameQueueFlag,
    NameReplaceExistingFlag,
    sd_bus_open,
)

if TYPE_CHECKING:
    from typing import Optional

    from .sd_bus_internals import SdBus

logger = getLogger(__name__)


class DefaultBusTLStorage(threading.local):
    bus: Optional[SdBus] = None


bus_tls = DefaultBusTLStorage()
bus_contextvar: ContextVar[SdBus] = ContextVar("DEFAULT_BUS")


def _get_defaul_bus_tls() -> Optional[SdBus]:
    return bus_tls.bus


def _set_default_bus_tls(new_bus: Optional[SdBus]) -> None:
    bus_tls.bus = new_bus


def get_default_bus() -> SdBus:
    """Get default bus.

    Returns context-local default bus if set or
    thread-local otherwise.

    If no default bus is set initializes a new bus using
    :py:func:`sdbus.sd_bus_open` and sets it as a thread-local
    default bus.
    """
    if (context_bus := bus_contextvar.get(None)) is not None:
        return context_bus

    if (tls_bus := _get_defaul_bus_tls()) is not None:
        return tls_bus
    else:
        new_bus = sd_bus_open()
        logger.info(
            "Created new default bus for thread %r",
            threading.current_thread(),
        )
        _set_default_bus_tls(new_bus)
        return new_bus


def set_default_bus(new_default: SdBus) -> None:
    """Set thread-local default bus.

    Should be called before creating any objects that will use
    default bus.

    Default bus can be replaced but the change will only affect
    newly created objects.
    """
    _set_default_bus_tls(new_default)


def set_context_default_bus(new_default: SdBus) -> Token[SdBus]:
    """Set context-local default bus.

    Should be called before creating any objects that will use
    default bus.

    Default bus can be replaced but the change will only affect
    newly created objects.

    Context-local default bus has higher priority over thread-local one
    but has to be explicitly set.

    :returns:
        Token that can be used to reset context bus back.
        See ``contextvars`` documentation for details.
    """
    return bus_contextvar.set(new_default)


def _prepare_request_name_flags(
        allow_replacement: bool,
        replace_existing: bool,
        queue: bool,
) -> int:
    return (
        (NameAllowReplacementFlag if allow_replacement else 0)
        +
        (NameReplaceExistingFlag if replace_existing else 0)
        +
        (NameQueueFlag if queue else 0)
    )


async def request_default_bus_name_async(
    new_name: str,
    allow_replacement: bool = False,
    replace_existing: bool = False,
    queue: bool = False,
) -> None:
    r"""Asynchronously acquire a name on the default bus.

    :param new_name:
        Name to acquire.
        Must be a valid D-Bus service name.
    :param allow_replacement:
        If name was acquired allow other D-Bus peers to take away the name.
    :param replace_existing:
        If current name owner allows, take away the name.
    :param queue:
        Queue up for name acquisition. :py:exc:`.SdBusRequestNameInQueueError`
        will be raised when successfully placed in queue. :py:meth:`Ownership
        change signal <sdbus_async.dbus_daemon.FreedesktopDbus.\
        name_owner_changed>` should be monitored get notified when the name
        was acquired.
    :raises: :ref:`name-request-exceptions` and other D-Bus exceptions.
    """
    default_bus = get_default_bus()
    await default_bus.request_name_async(
        new_name,
        _prepare_request_name_flags(
            allow_replacement,
            replace_existing,
            queue,
        )
    )


def request_default_bus_name(
    new_name: str,
    allow_replacement: bool = False,
    replace_existing: bool = False,
    queue: bool = False,
) -> None:
    r"""Acquire a name on the default bus.

    Blocks until a reply is received from D-Bus daemon.

    :param new_name:
        Name to acquire.
        Must be a valid D-Bus service name.
    :param allow_replacement:
        If name was acquired allow other D-Bus peers to take away the name.
    :param replace_existing:
        If current name owner allows, take away the name.
    :param queue:
        Queue up for name acquisition. :py:exc:`.SdBusRequestNameInQueueError`
        will be raised when successfully placed in queue. :py:meth:`Ownership
        change signal <sdbus_async.dbus_daemon.FreedesktopDbus.\
        name_owner_changed>` should be monitored get notified when the name
        was acquired.
    :raises: :ref:`name-request-exceptions` and other D-Bus exceptions.
    """
    default_bus = get_default_bus()
    default_bus.request_name(
        new_name,
        _prepare_request_name_flags(
            allow_replacement,
            replace_existing,
            queue,
        )
    )


__all__ = (
    "get_default_bus",
    "set_default_bus",
    "set_context_default_bus",
    "request_default_bus_name_async",
    "request_default_bus_name",
)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from typing import TYPE_CHECKING

from ..dbus_common_funcs import _parse_properties_vardict
from ..dbus_proxy_async_interface_base import (
    DBUS_CLASS_TO_META,
    DBUS_INTERFACE_NAME_TO_CLASS,
    DbusInterfaceBaseAsync,
)
from ..dbus_proxy_sync_interface_base import DbusInterfaceBase

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any, Literal, Optional, Union

    from ..dbus_proxy_async_interfaces import DBUS_PROPERTIES_CHANGED_TYPING

    InterfacesBaseClasses = Union[DbusInterfaceBaseAsync, DbusInterfaceBase]
    InterfacesBaseTypes = type[InterfacesBaseClasses]
    InterfacesInputElements = Union[InterfacesBaseClasses, InterfacesBaseTypes]
    InterfacesInput = Union[
        InterfacesInputElements,
        Iterable[InterfacesInputElements],
    ]
    InterfacesToClassMap = dict[
        frozenset[str],
        InterfacesBaseTypes,
    ]
    OnUnknownMember = Literal['error', 'ignore', 'reuse']
    OnUnknownInterface = Literal['error', 'none']
    ParseGetManaged = dict[
        str,
        tuple[
            Optional[InterfacesBaseTypes],
            dict[str, Any],
        ],
    ]


def _interfaces_input_to_types(
    interfaces: InterfacesInput,
) -> tuple[InterfacesBaseTypes, ...]:
    if isinstance(
        interfaces,
        (DbusInterfaceBaseAsync, DbusInterfaceBase, type)
    ):
        return (
            interfaces if isinstance(interfaces, type) else type(interfaces),
        )
    else:
        return tuple(i if isinstance(i, type) else type(i) for i in interfaces)


def parse_properties_changed(
    interface: InterfacesInputElements,
    properties_changed_data: DBUS_PROPERTIES_CHANGED_TYPING,
    on_unknown_member: OnUnknownMember = 'error',
) -> dict[str, Any]:
    """Parse data from :py:meth:`properties_changed \
    <sdbus.DbusInterfaceCommonAsync.properties_changed>` signal.

    Parses changed properties from a single D-Bus object. The object's
    interface class must be known in advance and passed as a first
    argument.

    Member names will be translated to python defined names.
    Invalidated properties will have a value of None.

    :param interface:
        Takes either D-Bus interface class or its object.
    :param properties_changed_data:
        Tuple caught from signal.
    :param on_unknown_member:
        If an unknown D-Bus property was encountered either raise
        an ``"error"`` (default), ``"ignore"`` the property
        or ``"reuse"`` the D-Bus name for the member.
    :returns:
        Dictionary of changed properties with keys translated to python
        names. Invalidated properties will have value of None.
    """
    interface_name, changed_properties, invalidated_properties = (
        properties_changed_data
    )

    meta = DBUS_CLASS_TO_META[DBUS_INTERFACE_NAME_TO_CLASS[interface_name]]

    for invalidated_property in invalidated_properties:
        changed_properties[invalidated_property] = ('0', None)

    return _parse_properties_vardict(
        meta.dbus_member_to_python_attr,
        properties_changed_data[1],
        on_unknown_member,
    )


SKIP_INTERFACES = frozenset((
    'org.freedesktop.DBus.Properties',
    'org.freedesktop.DBus.Introspectable',
    'org.freedesktop.DBus.Peer',
    'org.freedesktop.DBus.ObjectManager',
))


def _create_interfaces_map(
    interfaces: tuple[InterfacesBaseTypes, ...],
) -> InterfacesToClassMap:

    interfaces_to_class_map: InterfacesToClassMap = {}

    for interface in interfaces:
        interface_names_set = frozenset(
            interface_name for interface_name, _ in
            interface._dbus_iter_interfaces_meta()
            if interface_name not in SKIP_INTERFACES
        )
        interfaces_to_class_map[interface_names_set] = (
            interface if isinstance(interface, type)
            else type(interface)
        )

    return interfaces_to_class_map


def _get_class_from_interfaces(
    interfaces_to_class_map: InterfacesToClassMap,
    interface_names_iter: Iterable[str],
    raise_key_error: bool,
    use_subset: bool,
) -> Optional[InterfacesBaseTypes]:
    class_set = frozenset(interface_names_iter) - SKIP_INTERFACES
    if use_subset:
        for interface_available in sorted(
            interfaces_to_class_map.keys(),
            key=len,
            reverse=True,
        ):
            if interface_available.issubset(class_set):
                class_set = interface_available
                break

    try:
        return interfaces_to_class_map[class_set]
    except KeyError:
        if raise_key_error:
            raise

        return None


def _get_member_map_from_class(
    python_class: Optional[InterfacesBaseTypes],
) -> dict[str, dict[str, str]]:
    if python_class is None:
        return {}
    else:
        return {
            interface_name: meta.dbus_member_to_python_attr
            for interface_name, meta in
            python_class._dbus_iter_interfaces_meta()
        }


def _translate_and_merge_members(
    properties_data: dict[str, dict[str, Any]],
    dbus_to_python_map: dict[str, dict[str, str]],
    on_unknown_member: OnUnknownMember,
) -> dict[str, Any]:
    python_properties: dict[str, Any] = {}
    for interface_name, properties in properties_data.items():
        interface_member_map = dbus_to_python_map.get(
            interface_name, {},
        )
        python_properties.update(
            _parse_properties_vardict(
                interface_member_map,
                properties,
                on_unknown_member,
            )
        )

    return python_properties


def parse_interfaces_added(
    interfaces: InterfacesInput,
    interfaces_added_data: tuple[str, dict[str, dict[str, Any]]],
    on_unknown_interface: OnUnknownInterface = 'error',
    on_unknown_member: OnUnknownMember = 'error',
    *,
    use_interface_subsets: bool = False,
) -> tuple[str, Optional[InterfacesBaseTypes], dict[str, Any]]:
    """Parse data from :py:meth:`interfaces_added \
    <sdbus.DbusObjectManagerInterfaceAsync.interfaces_added>` signal.

    Takes the possible interface classes and the signal data.
    Returns the path of new object, the class of the
    added object (if it matched one of passed interface classes)
    and the dictionary of python named properties and their values.

    The passed interfaces can be async or blocking, the class
    or an instantiated object, a single item or an iterable of interfaces.

    :param interfaces:
        Possible interfaces that were added.
    :param interfaces_added_data:
        Tuple caught from signal.
    :param on_unknown_interface:
        If an unknown D-Bus interface was encountered either raise
        an ``"error"`` (default) or return ``"none"`` instead of
        interface class.
    :param on_unknown_member:
        If an unknown D-Bus property was encountered either raise
        an ``"error"`` (default), ``"ignore"`` the property
        or ``"reuse"`` the D-Bus name for the member.
    :param use_interface_subsets:
        Use the subset of interfaces as a valid match. For example,
        the class that implements ``org.example.foo`` would be matched
        with an data consising of both ``org.example.foo`` and
        ``org.example.bar``. The classes implementing more interfaces
        will have higher priority over the ones implementing fewer.
    :returns:
        Path of new added object, object's class (or ``None``) and dictionary
        of python translated members and their values.
    """
    interfaces_types = _interfaces_input_to_types(interfaces)
    interfaces_to_class_map = _create_interfaces_map(interfaces_types)

    path, properties_data = interfaces_added_data

    python_class = (
        _get_class_from_interfaces(
            interfaces_to_class_map,
            properties_data.keys(),
            on_unknown_interface == "error",
            use_interface_subsets,
        )
    )
    dbus_to_python_member_map = _get_member_map_from_class(python_class)
    python_properties: dict[str, Any] = {}
    for interface_name, properties in properties_data.items():
        interface_member_map = dbus_to_python_member_map.get(
            interface_name, {},
        )
        python_properties.update(
            _parse_properties_vardict(
                interface_member_map,
                properties,
                on_unknown_member,
            )
        )

    return (
        path,
        python_class,
        _translate_and_merge_members(
            properties_data,
            dbus_to_python_member_map,
            on_unknown_member,
        ),
    )


def parse_interfaces_removed(
    interfaces: InterfacesInput,
    interfaces_removed_data: tuple[str, list[str]],
    on_unknown_interface: OnUnknownInterface = 'error',
    *,
    use_interface_subsets: bool = False,
) -> tuple[str, Optional[InterfacesBaseTypes]]:
    """Parse data from :py:meth:`interfaces_added \
    <sdbus.DbusObjectManagerInterfaceAsync.interfaces_removed>` signal.

    Takes the possible interface classes and the signal data.
    Returns the path and the matched class of removed object.
    (if it matched one of passed interface classes)

    The passed interfaces can be async or blocking, the class
    or an instantiated object, a single item or an iterable of interfaces.

    :param interfaces:
        Possible interfaces that were removed.
    :param interfaces_added_data:
        Tuple caught from signal.
    :param on_unknown_member:
        If an unknown D-Bus interface was encountered either raise an
        ``"error"`` (default) or return ``"none"`` instead of interface class.
    :param use_interface_subsets:
        Use the subset of interfaces as a valid match. For example,
        the class that implements ``org.example.foo`` would be matched
        with an data consising of both ``org.example.foo`` and
        ``org.example.bar``. The classes implementing more interfaces
        will have higher priority over the ones implementing fewer.
    :returns:
        Path of removed object and object's class (or ``None``).
    """
    interfaces_types = _interfaces_input_to_types(interfaces)
    interfaces_to_class_map = _create_interfaces_map(interfaces_types)

    path, interfaces_removed = interfaces_removed_data

    python_class = (
        _get_class_from_interfaces(
            interfaces_to_class_map,
            interfaces_removed,
            on_unknown_interface == "error",
            use_interface_subsets,
        )
    )

    return path, python_class


def parse_get_managed_objects(
    interfaces: InterfacesInput,
    managed_objects_data: dict[str, dict[str, dict[str, Any]]],
    on_unknown_interface: OnUnknownInterface = 'error',
    on_unknown_member: OnUnknownMember = 'error',
    *,
    use_interface_subsets: bool = False,
) -> ParseGetManaged:
    """Parse data from :py:meth:`get_managed_objects \
    <sdbus.DbusObjectManagerInterfaceAsync.get_managed_objects>` call.

    Takes the possible interface classes and the method's returned data.
    Returns a dictionary where keys a paths of the managed objects and
    value is a tuple of class of the object and dictionary of its python
    named properties and their values.

    The passed interfaces can be async or blocking, the class
    or an instantiated object, a single item or an iterable of interfaces.

    :param interfaces:
        Possible interfaces of the managed objects.
    :param managed_objects_data:
        Data returned by ``get_managed_objects`` call.
    :param on_unknown_interface:
        If an unknown D-Bus interface was encountered either raise an
        ``"error"`` (default) or return ``"none"`` instead of interface class.
    :param on_unknown_member:
        If an unknown D-Bus property was encountered either raise
        an ``"error"`` (default), ``"ignore"`` the property
        or ``"reuse"`` the D-Bus name for the member.
    :param use_interface_subsets:
        Use the subset of interfaces as a valid match. For example,
        the class that implements ``org.example.foo`` would be matched
        with an data consising of both ``org.example.foo`` and
        ``org.example.bar``. The classes implementing more interfaces
        will have higher priority over the ones implementing fewer.
    :returns:
        Dictionary where keys are paths and values are tuples of managed
        objects classes and their properties data.

    *New in version 0.12.0.*
    """
    interfaces_types = _interfaces_input_to_types(interfaces)
    interfaces_to_class_map = _create_interfaces_map(interfaces_types)

    managed_objects_map: ParseGetManaged = {}

    for path, properties_data in managed_objects_data.items():
        python_class = (
            _get_class_from_interfaces(
                interfaces_to_class_map,
                properties_data.keys(),
                on_unknown_interface == "error",
                use_interface_subsets,
            )
        )
        dbus_to_python_member_map = _get_member_map_from_class(python_class)

        managed_objects_map[path] = (
            python_class,
            _translate_and_merge_members(
                properties_data,
                dbus_to_python_member_map,
                on_unknown_member,
            ),
        )

    return managed_objects_map


__all__ = (
    'parse_properties_changed',
    'parse_interfaces_added',
    'parse_interfaces_removed',
    'parse_get_managed_objects',
)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from .dbus_exceptions import (
    DbusAccessDeniedError,
    DbusAddressInUseError,
    DbusAuthFailedError,
    DbusBadAddressError,
    DbusDisconnectedError,
    DbusFailedError,
    DbusFileExistsError,
    DbusFileNotFoundError,
    DbusInconsistentMessageError,
    DbusInteractiveAuthorizationRequiredError,
    DbusInvalidArgsError,
    DbusInvalidFileContentError,
    DbusInvalidSignatureError,
    DbusIOError,
    DbusLimitsExceededError,
    DbusMatchRuleInvalidError,
    DbusMatchRuleNotFound,
    DbusNameHasNoOwnerError,
    DbusNoMemoryError,
    DbusNoNetworkError,
    DbusNoReplyError,
    DbusNoServerError,
    DbusNotSupportedError,
    DbusPropertyReadOnlyError,
    DbusServiceUnknownError,
    DbusTimeoutError,
    DbusUnixProcessIdUnknownError,
    DbusUnknownInterfaceError,
    DbusUnknownMethodError,
    DbusUnknownObjectError,
    DbusUnknownPropertyError,
)
from .sd_bus_internals import (
    SdBusBaseError,
    SdBusLibraryError,
    SdBusRequestNameAlreadyOwnerError,
    SdBusRequestNameError,
    SdBusRequestNameExistsError,
    SdBusRequestNameInQueueError,
    SdBusUnmappedMessageError,
    map_exception_to_dbus_error,
)

__all__ = (
    'DbusAccessDeniedError',
    'DbusAddressInUseError',
    'DbusAuthFailedError',
    'DbusBadAddressError',
    'DbusDisconnectedError',
    'DbusFailedError',
    'DbusFileExistsError',
    'DbusFileNotFoundError',
    'DbusInconsistentMessageError',
    'DbusInteractiveAuthorizationRequiredError',
    'DbusInvalidArgsError',
    'DbusInvalidFileContentError',
    'DbusInvalidSignatureError',
    'DbusIOError',
    'DbusLimitsExceededError',
    'DbusMatchRuleInvalidError',
    'DbusMatchRuleNotFound',
    'DbusNameHasNoOwnerError',
    'DbusNoMemoryError',
    'DbusNoNetworkError',
    'DbusNoReplyError',
    'DbusNoServerError',
    'DbusNotSupportedError',
    'DbusPropertyReadOnlyError',
    'DbusServiceUnknownError',
    'DbusTimeoutError',
    'DbusUnixProcessIdUnknownError',
    'DbusUnknownInterfaceError',
    'DbusUnknownMethodError',
    'DbusUnknownObjectError',
    'DbusUnknownPropertyError',
    'map_exception_to_dbus_error',

    'SdBusBaseError',
    'SdBusLibraryError',
    'SdBusRequestNameAlreadyOwnerError',
    'SdBusRequestNameError',
    'SdBusRequestNameExistsError',
    'SdBusRequestNameInQueueError',
    'SdBusUnmappedMessageError',
)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary, WeakValueDictionary

from .dbus_common_elements import (
    DbusClassMeta,
    DbusInterfaceMetaCommon,
    DbusMemberAsync,
    DbusMemberCommon,
    DbusRemoteObjectMeta,
)
from .dbus_proxy_sync_method import DbusMethodSync
from .dbus_proxy_sync_property import DbusPropertySync

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Any, Optional

    from .sd_bus_internals import SdBus


DBUS_CLASS_TO_META: WeakKeyDictionary[
    type, DbusClassMeta] = WeakKeyDictionary()
DBUS_INTERFACE_NAME_TO_CLASS: WeakValueDictionary[
    str, DbusInterfaceMetaSync] = WeakValueDictionary()


class DbusInterfaceMetaSync(DbusInterfaceMetaCommon):

    @staticmethod
    def _check_collisions(
        new_class_name: str,
        attr_names: set[str],
        reserved_attr_names: set[str],
    ) -> None:

        possible_collisions = attr_names & reserved_attr_names
        if possible_collisions:
            raise ValueError(
                f"Interface {new_class_name!r} redefines reserved "
                f"D-Bus attribute names: {possible_collisions!r}"
            )

    @staticmethod
    def _collect_dbus_to_python_attr_names(
        new_class_name: str,
        base_classes: Iterable[type],
    ) -> set[str]:
        all_python_dbus_attrs: set[str] = set()
        possible_collisions: set[str] = set()

        for c in base_classes:
            dbus_meta = DBUS_CLASS_TO_META.get(c)
            if dbus_meta is None:
                continue

            base_python_dbus_attrs = set(
                dbus_meta.python_attr_to_dbus_member.keys()
            )

            possible_collisions.update(
                base_python_dbus_attrs & all_python_dbus_attrs
            )

            all_python_dbus_attrs.update(
                base_python_dbus_attrs
            )

        if possible_collisions:
            raise ValueError(
                f"Interface {new_class_name!r} has a reserved D-Bus "
                f"attribute name collision: {possible_collisions!r}"
            )

        return all_python_dbus_attrs

    @staticmethod
    def _map_dbus_elements(
        attr_name: str,
        attr: Any,
        meta: DbusClassMeta,
    ) -> None:
        if not isinstance(attr, DbusMemberCommon):
            return

        if isinstance(attr, DbusMemberAsync):
            raise TypeError(
                f"Can't mix async methods in sync interface: {attr_name!r}"
            )

        if isinstance(attr, DbusMethodSync):
            meta.dbus_member_to_python_attr[attr.method_name] = attr_name
            meta.python_attr_to_dbus_member[attr_name] = attr.method_name
        elif isinstance(attr, DbusPropertySync):
            meta.dbus_member_to_python_attr[attr.property_name] = attr_name
            meta.python_attr_to_dbus_member[attr_name] = attr.property_name
        else:
            raise TypeError(f"Unknown D-Bus element: {attr!r}")

    def __new__(cls, name: str,
                bases: tuple[type, ...],
                namespace: dict[str, Any],
                interface_name: Optional[str] = None,
                serving_enabled: bool = True,
                ) -> DbusInterfaceMetaSync:

        if interface_name in DBUS_INTERFACE_NAME_TO_CLASS:
            raise ValueError(
                f"D-Bus interface of the name {interface_name!r} was "
                "already created."
            )

        all_mro_bases: set[type[Any]] = set(
            chain.from_iterable(c.__mro__ for c in bases)
        )
        reserved_attr_names = cls._collect_dbus_to_python_attr_names(
            name, all_mro_bases,
        )
        cls._check_collisions(name, set(namespace.keys()), reserved_attr_names)

        new_cls = super().__new__(
            cls, name, bases, namespace,
            interface_name,
            serving_enabled,
        )

        if interface_name is not None:
            dbus_class_meta = DbusClassMeta(interface_name, serving_enabled)
            DBUS_CLASS_TO_META[new_cls] = dbus_class_meta
            DBUS_INTERFACE_NAME_TO_CLASS[interface_name] = new_cls

            for attr_name, attr in namespace.items():
                cls._map_dbus_elements(attr_name, attr, dbus_class_meta)

        return new_cls


class DbusInterfaceBase(metaclass=DbusInterfaceMetaSync):

    def __init__(
        self,
        service_name: str,
        object_path: str,
        bus: Optional[SdBus] = None,
    ):
        self._dbus = DbusRemoteObjectMeta(service_name, object_path, bus)

    @classmethod
    def _dbus_iter_interfaces_meta(
        cls,
    ) -> Iterator[tuple[str, DbusClassMeta]]:

        for base in reversed(cls.__mro__):
            meta = DBUS_CLASS_TO_META.get(base)
            if meta is None:
                continue

            yield meta.interface_name, meta



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from sdbus.unittest import IsolatedDbusTestCase

from sdbus import DbusInterfaceCommonAsync, dbus_method_async

TEST_SERVICE_NAME = 'org.example.test'


def initialize_object(
    interface_class: type[DbusInterfaceCommonAsync],
) -> tuple[DbusInterfaceCommonAsync, DbusInterfaceCommonAsync]:
    test_object = interface_class()
    test_object.export_to_dbus('/')

    test_object_connection = interface_class.new_proxy(
        TEST_SERVICE_NAME, '/')

    return test_object, test_object_connection


class TestIntrospection(IsolatedDbusTestCase):

    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        await self.bus.request_name_async("org.example.test", 0)

    async def test_method_arg_names_none(self) -> None:
        class TestInterface(
            DbusInterfaceCommonAsync,
            interface_name="org.test.intro1",
        ):
            @dbus_method_async(
                input_signature="ss",
                result_signature="i",
            )
            async def login(
                self,
                user_name: str,
                pin_code: str,
            ) -> int:
                return 0

        obj, rem = initialize_object(TestInterface)

        introspection = await rem.dbus_introspect()
        self.assertNotIn('name="user_name"', introspection)
        self.assertNotIn('name="result"', introspection)
        self.assertNotIn('name="pin_code"', introspection)

    async def test_method_arg_names_result_names_only(self) -> None:
        class TestInterface(
            DbusInterfaceCommonAsync,
            interface_name="org.test.intro2",
        ):
            @dbus_method_async(
                input_signature="ss",
                result_signature="i",
                result_args_names=("result",)
            )
            async def login(
                self,
                user_name: str,
                pin_code: str,
            ) -> int:
                return 0

        obj, rem = initialize_object(TestInterface)

        introspection = await rem.dbus_introspect()
        self.assertIn('name="user_name"', introspection)
        self.assertIn('name="result"', introspection)
        self.assertIn('name="pin_code"', introspection)

    async def test_method_arg_names_full(self) -> None:
        class TestInterface(
            DbusInterfaceCommonAsync,
            interface_name="org.test.intro3",
        ):
            @dbus_method_async(
                input_signature="ss",
                input_args_names=("UserName", "PinCode"),
                result_signature="i",
                result_args_names=("Result",)
            )
            async def login(
                self,
                user_name: str,
                pin_code: str,
            ) -> int:
                return 0

        obj, rem = initialize_object(TestInterface)

        introspection = await rem.dbus_introspect()
        self.assertIn('name="UserName"', introspection)
        self.assertIn('name="Result"', introspection)
        self.assertIn('name="PinCode"', introspection)

    async def test_method_arg_names_no_return_args(self) -> None:
        class TestInterface(
            DbusInterfaceCommonAsync,
            interface_name="org.test.intro4",
        ):
            @dbus_method_async(
                input_signature="ss",
                result_args_names=(),
            )
            async def login(
                self,
                user_name: str,
                pin_code: str,
            ) -> None:
                return None

        obj, rem = initialize_object(TestInterface)

        introspection = await rem.dbus_introspect()
        self.assertIn('name="user_name"', introspection)
        self.assertIn('name="pin_code"', introspection)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from asyncio import get_running_loop
from unittest import SkipTest, TestCase, main

from sdbus.sd_bus_internals import (
    SdBus,
    is_interface_name_valid,
    is_member_name_valid,
    is_object_path_valid,
    is_service_name_valid,
)
from sdbus.unittest import IsolatedDbusTestCase


class TestAsyncLowLevel(IsolatedDbusTestCase):
    def test_init_bus(self) -> None:
        not_connected_bus = SdBus()
        self.assertIsNone(not_connected_bus.address)

        self.assertIsNotNone(self.bus.address)

    async def test_bus_fd_unregister_close(self) -> None:
        await self.bus.request_name_async("org.example", 0)
        bus_fd = self.bus.get_fd()

        self.bus.close()

        loop = get_running_loop()
        self.assertFalse(loop.remove_reader(bus_fd))
        self.assertFalse(loop.remove_writer(bus_fd))


class TestLowLeveApi(TestCase):
    def test_validation_funcs(self) -> None:
        try:
            self.assertTrue(
                is_interface_name_valid('org.example.test')
            )

            self.assertFalse(
                is_interface_name_valid('Not very valid 😀')
            )

            self.assertTrue(
                is_service_name_valid('org.example.test')
            )

            self.assertFalse(
                is_service_name_valid('Not very valid 😀')
            )

            self.assertTrue(
                is_member_name_valid('GetSomething')
            )

            self.assertFalse(
                is_member_name_valid('no.dots.in.member.names')
            )

            self.assertTrue(
                is_object_path_valid('/test')
            )

            self.assertFalse(
                is_object_path_valid('no.dots.in.object.paths')
            )
        except NotImplementedError:
            raise SkipTest(
                "Validation funcs not implemented. "
                "Probably too old libsystemd. (< 246)"
            )

    def test_bus_method_call_timeout(self) -> None:
        bus = SdBus()

        self.assertIsNotNone(bus.method_call_timeout_usec)

        test_timeout_usec = 10 * 10**6  # 10 seconds
        bus.method_call_timeout_usec = test_timeout_usec
        self.assertEqual(test_timeout_usec, bus.method_call_timeout_usec)

        with self.assertRaises(TypeError):
            bus.method_call_timeout_usec = "test"  # type: ignore

        with self.assertRaises(ValueError):
            del bus.method_call_timeout_usec


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from inspect import iscoroutinefunction
from types import FunctionType
from typing import TYPE_CHECKING, Generic, TypeVar, cast

from .dbus_common_elements import DbusMemberSync, DbusPropertyCommon
from .dbus_common_funcs import _check_sync_in_async_env

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Optional

    from .dbus_proxy_sync_interface_base import DbusInterfaceBase


T = TypeVar('T')


class DbusPropertySync(DbusPropertyCommon, DbusMemberSync, Generic[T]):
    def __init__(
            self,
            property_name: Optional[str],
            property_signature: str,
            property_getter: Callable[[DbusInterfaceBase],
                                      T],
            property_setter: Optional[
                Callable[[DbusInterfaceBase, T],
                         None]],
            flags: int,

    ) -> None:
        assert isinstance(property_getter, FunctionType)
        super().__init__(
            property_name,
            property_signature,
            flags,
            property_getter,
        )
        self.property_getter = property_getter
        self.property_setter = property_setter

        self.__doc__ = property_getter.__doc__

    def __get__(self,
                obj: DbusInterfaceBase,
                obj_class: Optional[type[DbusInterfaceBase]] = None,
                ) -> T:
        assert _check_sync_in_async_env(), (
            "Used sync __get__ method in async environment. "
            "This is probably an error as it will block "
            "other asyncio methods for considerable time."
        )

        new_call_message = (
            obj._dbus.attached_bus.new_property_get_message(
                obj._dbus.service_name,
                obj._dbus.object_path,
                self.interface_name,
                self.property_name,
            )
        )

        reply_message = obj._dbus.attached_bus.call(new_call_message)
        return cast(T, reply_message.get_contents()[1])

    def __set__(self, obj: DbusInterfaceBase, value: T) -> None:
        assert _check_sync_in_async_env(), (
            "Used sync __set__ method in async environment. "
            "This is probably an error as it will block "
            "other asyncio methods for considerable time."
        )

        if not self.property_signature:
            raise AttributeError('D-Bus property is read only')

        new_call_message = (
            obj._dbus.attached_bus.new_property_set_message(
                obj._dbus.service_name,
                obj._dbus.object_path,
                self.interface_name,
                self.property_name,
            )
        )

        new_call_message.append_data(
            'v', (self.property_signature, value))

        obj._dbus.attached_bus.call(new_call_message)


def dbus_property(
    property_signature: str = "",
    flags: int = 0,
    property_name: Optional[str] = None,
) -> Callable[
    [Callable[[Any], T]],
        DbusPropertySync[T]]:
    assert not isinstance(property_signature, FunctionType), (
        "Passed function to decorator directly. "
        "Did you forget () round brackets?"
    )

    def property_decorator(
        function: Callable[..., Any]
    ) -> DbusPropertySync[T]:
        assert not iscoroutinefunction(function), "Expected regular function"

        new_wrapper: DbusPropertySync[T] = DbusPropertySync(
            property_name,
            property_signature,
            function,
            None,
            flags,
        )

        return new_wrapper

    return property_decorator



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020, 2021 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from typing import TYPE_CHECKING

from .sd_bus_internals import (
    SdBusBaseError,
    add_exception_mapping,
    map_exception_to_dbus_error,
)

if TYPE_CHECKING:
    from typing import Any


class DbusErrorMeta(type):

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
    ) -> type[Exception]:

        dbus_error_name = namespace.get('dbus_error_name')

        if dbus_error_name is None:
            raise TypeError('D-Bus error name not passed')

        new_cls = super().__new__(cls, name, bases, namespace)
        assert issubclass(new_cls, Exception), (
            f"New class {new_cls} is not an Exception but {bases}."
        )

        add_exception_mapping(new_cls)

        return new_cls


class DbusFailedError(SdBusBaseError, metaclass=DbusErrorMeta):
    dbus_error_name: str = 'org.freedesktop.DBus.Error.Failed'


class DbusNoMemoryError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.NoMemory'


class DbusServiceUnknownError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.ServiceUnknown'


class DbusNameHasNoOwnerError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.NameHasNoOwner'


class DbusNoReplyError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.NoReply'


class DbusIOError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.IOError'


class DbusBadAddressError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.BadAddress'


class DbusNotSupportedError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.NotSupported'


class DbusLimitsExceededError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.LimitsExceeded'


class DbusAccessDeniedError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.AccessDenied'


class DbusAuthFailedError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.AuthFailed'


class DbusNoServerError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.NoServer'


class DbusTimeoutError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.Timeout'


class DbusNoNetworkError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.NoNetwork'


class DbusAddressInUseError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.AddressInUse'


class DbusDisconnectedError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.Disconnected'


class DbusInvalidArgsError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.InvalidArgs'


class DbusFileNotFoundError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.FileNotFound'


class DbusFileExistsError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.FileExists'


class DbusUnknownMethodError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.UnknownMethod'


class DbusUnknownObjectError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.UnknownObject'


class DbusUnknownInterfaceError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.UnknownInterface'


class DbusUnknownPropertyError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.UnknownProperty'


class DbusPropertyReadOnlyError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.PropertyReadOnly'


class DbusUnixProcessIdUnknownError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.UnixProcessIdUnknown'


class DbusInvalidSignatureError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.InvalidSignature'


class DbusInvalidFileContentError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.InvalidFileContent'


class DbusInconsistentMessageError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.InconsistentMessage'


class DbusMatchRuleNotFound(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.MatchRuleNotFound'


class DbusMatchRuleInvalidError(DbusFailedError):
    dbus_error_name = 'org.freedesktop.DBus.Error.MatchRuleInvalid'


class DbusInteractiveAuthorizationRequiredError(DbusFailedError):
    dbus_error_name = ('org.freedesktop.DBus.Error'
                       '.InteractiveAuthorizationRequired')


map_exception_to_dbus_error(AssertionError,
                            "org.python.Error.AssertionError")
map_exception_to_dbus_error(AttributeError,
                            "org.python.Error.AttributeError")
map_exception_to_dbus_error(BufferError,
                            "org.python.Error.BufferError")
map_exception_to_dbus_error(EOFError,
                            "org.python.Error.EOFError")
map_exception_to_dbus_error(ImportError,
                            "org.python.Error.ImportError")
map_exception_to_dbus_error(ModuleNotFoundError,
                            "org.python.Error.ModuleNotFoundError")
map_exception_to_dbus_error(LookupError,
                            "org.python.Error.LookupError")
map_exception_to_dbus_error(IndexError,
                            "org.python.Error.IndexError")
map_exception_to_dbus_error(KeyError,
                            "org.python.Error.KeyError")
map_exception_to_dbus_error(NameError,
                            "org.python.Error.NameError")
map_exception_to_dbus_error(NotImplementedError,
                            "org.python.Error.NotImplementedError")
map_exception_to_dbus_error(OSError,
                            "org.python.Error.OSError")
map_exception_to_dbus_error(RecursionError,
                            "org.python.Error.RecursionError")
map_exception_to_dbus_error(ReferenceError,
                            "org.python.Error.ReferenceError")
map_exception_to_dbus_error(RuntimeError,
                            "org.python.Error.RuntimeError")
map_exception_to_dbus_error(SyntaxError,
                            "org.python.Error.SyntaxError")
map_exception_to_dbus_error(IndentationError,
                            "org.python.Error.IndentationError")
map_exception_to_dbus_error(TabError,
                            "org.python.Error.TabError")
map_exception_to_dbus_error(SystemError,
                            "org.python.Error.SystemError")
map_exception_to_dbus_error(TypeError,
                            "org.python.Error.TypeError")
map_exception_to_dbus_error(UnboundLocalError,
                            "org.python.Error.UnboundLocalError")
map_exception_to_dbus_error(UnicodeError,
                            "org.python.Error.UnicodeError")
map_exception_to_dbus_error(UnicodeEncodeError,
                            "org.python.Error.UnicodeEncodeError")
map_exception_to_dbus_error(UnicodeDecodeError,
                            "org.python.Error.UnicodeDecodeError")
map_exception_to_dbus_error(UnicodeTranslateError,
                            "org.python.Error.UnicodeTranslateError")
map_exception_to_dbus_error(ValueError,
                            "org.python.Error.ValueError")
map_exception_to_dbus_error(EnvironmentError,
                            "org.python.Error.EnvironmentError")
map_exception_to_dbus_error(IOError,
                            "org.python.Error.IOError")
map_exception_to_dbus_error(BlockingIOError,
                            "org.python.Error.BlockingIOError")
map_exception_to_dbus_error(ChildProcessError,
                            "org.python.Error.ChildProcessError")
map_exception_to_dbus_error(ConnectionError,
                            "org.python.Error.ConnectionError")
map_exception_to_dbus_error(BrokenPipeError,
                            "org.python.Error.BrokenPipeError")
map_exception_to_dbus_error(ConnectionAbortedError,
                            "org.python.Error.ConnectionAbortedError")
map_exception_to_dbus_error(ConnectionRefusedError,
                            "org.python.Error.ConnectionRefusedError")
map_exception_to_dbus_error(ConnectionResetError,
                            "org.python.Error.ConnectionResetError")
map_exception_to_dbus_error(FileExistsError,
                            "org.python.Error.FileExistsError")
map_exception_to_dbus_error(FileNotFoundError,
                            "org.python.Error.FileNotFoundError")
map_exception_to_dbus_error(InterruptedError,
                            "org.python.Error.InterruptedError")
map_exception_to_dbus_error(IsADirectoryError,
                            "org.python.Error.IsADirectoryError")
map_exception_to_dbus_error(NotADirectoryError,
                            "org.python.Error.NotADirectoryError")
map_exception_to_dbus_error(PermissionError,
                            "org.python.Error.PermissionError")
map_exception_to_dbus_error(ProcessLookupError,
                            "org.python.Error.ProcessLookupError")
map_exception_to_dbus_error(TimeoutError,
                            "org.python.Error.TimeoutError")
map_exception_to_dbus_error(ArithmeticError,
                            "org.python.Error.ArithmeticError")
map_exception_to_dbus_error(FloatingPointError,
                            "org.python.Error.FloatingPointError")
map_exception_to_dbus_error(OverflowError,
                            "org.python.Error.OverflowError")
map_exception_to_dbus_error(ZeroDivisionError,
                            "org.python.Error.ZeroDivisionError")



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2024 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from .dbus_common_elements import DbusLocalObjectMeta
from .dbus_proxy_async_interface_base import (
    DbusExportHandle,
    DbusInterfaceBaseAsync,
)
from .dbus_proxy_async_interfaces import DbusInterfaceCommonAsync
from .dbus_proxy_async_method import dbus_method_async
from .dbus_proxy_async_signal import dbus_signal_async
from .default_bus import get_default_bus

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Optional

    from .sd_bus_internals import SdBus, SdBusSlot


class DbusObjectManagerExportHandle(DbusExportHandle):
    def __init__(
        self,
        local_meta: DbusLocalObjectMeta,
        remove_object_call: Callable[[], None],
    ):
        super().__init__(local_meta)
        self.remove_object_call = remove_object_call

    def stop(self) -> None:
        self.remove_object_call()
        super().stop()


class DbusObjectManagerInterfaceAsync(
    DbusInterfaceCommonAsync,
    interface_name='org.freedesktop.DBus.ObjectManager',
    serving_enabled=False,
):
    def __init__(self) -> None:
        super().__init__()
        self._object_manager_slot: Optional[SdBusSlot] = None
        self._managed_object_to_path: dict[DbusInterfaceBaseAsync, str] = {}

    @dbus_method_async(result_signature='a{oa{sa{sv}}}')
    async def get_managed_objects(
            self) -> dict[str, dict[str, dict[str, Any]]]:
        raise NotImplementedError

    @dbus_signal_async('oa{sa{sv}}')
    def interfaces_added(self) -> tuple[str, dict[str, dict[str, Any]]]:
        raise NotImplementedError

    @dbus_signal_async('oao')
    def interfaces_removed(self) -> tuple[str, list[str]]:
        raise NotImplementedError

    def _dbus_on_no_members_exported(self) -> None:
        ...  # Object manager is allowed to be exported empty

    def export_to_dbus(
        self,
        object_path: str,
        bus: Optional[SdBus] = None,
    ) -> DbusExportHandle:
        if bus is None:
            bus = get_default_bus()

        export_handle = super().export_to_dbus(
            object_path,
            bus,
        )
        slot = bus.add_object_manager(object_path)
        self._object_manager_slot = slot
        export_handle._dbus_slots.append(slot)
        return export_handle

    def export_with_manager(
        self,
        object_path: str,
        object_to_export: DbusInterfaceBaseAsync,
        bus: Optional[SdBus] = None,
    ) -> DbusObjectManagerExportHandle:
        if self._object_manager_slot is None:
            raise RuntimeError('ObjectManager not intitialized')

        if bus is None:
            bus = get_default_bus()

        object_to_export.export_to_dbus(
            object_path,
            bus,
        )
        meta = object_to_export._dbus
        if not isinstance(meta, DbusLocalObjectMeta):
            raise TypeError
        handle = DbusObjectManagerExportHandle(
            meta,
            partial(self.remove_managed_object, object_to_export),
        )
        bus.emit_object_added(object_path)
        self._managed_object_to_path[object_to_export] = object_path

        return handle

    def remove_managed_object(
            self,
            managed_object: DbusInterfaceBaseAsync) -> None:
        if self._dbus.attached_bus is None:
            raise RuntimeError('Object manager not exported')

        removed_path = self._managed_object_to_path.pop(managed_object)
        self._dbus.attached_bus.emit_object_removed(removed_path)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020, 2021 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from os import environ
from subprocess import DEVNULL, PIPE
from subprocess import run as subprocess_run
from typing import Optional

from setuptools import Extension, setup

c_macros: list[tuple[str, Optional[str]]] = []


def get_libsystemd_version() -> int:
    process = subprocess_run(
        args=('pkg-config', '--modversion', 'libsystemd'),
        stderr=DEVNULL,
        stdout=PIPE,
        check=True,
        text=True,
    )

    result_str = process.stdout
    # Version can either be like 250 or 250.10
    first_component = result_str.split(".")[0]

    return int(first_component)


if not environ.get('PYTHON_SDBUS_USE_IGNORE_SYSTEMD_VERSION'):
    systemd_version = get_libsystemd_version()

    if systemd_version < 246:
        c_macros.append(('LIBSYSTEMD_NO_VALIDATION_FUNCS', None))

    if systemd_version < 248:
        c_macros.append(('LIBSYSTEMD_NO_OPEN_USER_MACHINE', None))


def get_link_arguments() -> list[str]:
    process = subprocess_run(
        args=('pkg-config', '--libs-only-l', 'libsystemd'),
        stderr=DEVNULL,
        stdout=PIPE,
        check=True,
    )

    result_str = process.stdout.decode('utf-8')

    return result_str.rstrip(' \n').split(' ')


link_arguments: list[str] = get_link_arguments()

if environ.get('PYTHON_SDBUS_USE_STATIC_LINK'):
    # Link statically against libsystemd and libcap
    link_arguments = ['-Wl,-Bstatic', *link_arguments, '-lcap',
                      '-Wl,-Bdynamic', '-lrt', '-lpthread']

link_arguments.append('-flto')

compile_arguments: list[str] = ['-flto']

use_limited_api = False

if environ.get('PYTHON_SDBUS_USE_LIMITED_API'):
    c_macros.append(('Py_LIMITED_API', '0x03090000'))
    use_limited_api = True


if __name__ == '__main__':
    with open('./README.md') as f:
        long_description = f.read()

    setup(
        name='sdbus',
        description=('Modern Python D-Bus library. '
                     'Based on sd-bus from libsystemd.'),
        long_description=long_description,
        long_description_content_type='text/markdown',
        version='0.14.1.post0',
        url='https://github.com/python-sdbus/python-sdbus',
        author='igo95862',
        author_email='igo95862@yandex.ru',
        license='LGPL-2.1-or-later',
        keywords='dbus ipc linux freedesktop',
        project_urls={
            'Documentation': 'https://python-sdbus.readthedocs.io/en/latest/',
            'Source': 'https://github.com/python-sdbus/python-sdbus/',
            'Tracker': 'https://github.com/python-sdbus/python-sdbus/issues/',
        },
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            (
                'License :: OSI Approved :: '
                'GNU Lesser General Public License v2 or later (LGPLv2+)'
            ),
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python :: 3 :: Only',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        packages=[
            'sdbus',
            'sdbus.utils',
            'sdbus_async.dbus_daemon',
            'sdbus_block.dbus_daemon',
        ],
        package_dir={
            'sdbus': 'src/sdbus',
            'sdbus_async.dbus_daemon': 'src/sdbus_async/dbus_daemon',
            'sdbus_block.dbus_daemon': 'src/sdbus_block/dbus_daemon',
        },
        package_data={
            'sdbus': [
                'py.typed',
                'sd_bus_internals.pyi',
                'sd_bus_internals.h',
            ],
            'sdbus_async.dbus_daemon': [
                'py.typed',
            ],
            'sdbus_block.dbus_daemon': [
                'py.typed',
            ],
        },
        python_requires='>=3.9',
        ext_modules=[
            Extension(
                'sdbus.sd_bus_internals',
                [
                    'src/sdbus/sd_bus_internals.c',
                    'src/sdbus/sd_bus_internals_bus.c',
                    'src/sdbus/sd_bus_internals_funcs.c',
                    'src/sdbus/sd_bus_internals_interface.c',
                    'src/sdbus/sd_bus_internals_message.c',
                ],
                extra_compile_args=compile_arguments,
                extra_link_args=link_arguments,
                define_macros=c_macros,
                py_limited_api=use_limited_api,
            )
        ],
    )



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020, 2021 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from unittest import main

from sdbus.exceptions import DbusPropertyReadOnlyError
from sdbus.unittest import IsolatedDbusTestCase
from sdbus_block.dbus_daemon import FreedesktopDbus

from sdbus import DbusInterfaceCommon, dbus_method


class TestSync(IsolatedDbusTestCase):

    def test_sync(self) -> None:
        self.bus.request_name('org.example.test', 0)

        s = FreedesktopDbus(self.bus)
        s.dbus_ping()
        s.dbus_introspect()
        s.dbus_machine_id()

        self.assertIsInstance(s.get_id(), str)
        self.assertIsInstance(s.features, list)
        self.assertIsInstance(
            s.get_connection_pid('org.freedesktop.DBus'), int)

        self.assertIsInstance(
            s.get_connection_uid('org.freedesktop.DBus'), int)

        with self.assertRaises(DbusPropertyReadOnlyError):
            s.features = ['test']

        self.assertTrue(s.get_name_owner('org.example.test'))

        with self.subTest('Test dbus to python name map'):
            self.assertTrue(
                any(
                    "Features" in meta.dbus_member_to_python_attr
                    for meta in (
                        meta for _, meta in s._dbus_iter_interfaces_meta()
                    )
                )
            )

        with self.subTest('Test properties_get_all_dict'):
            self.assertIn('features', s.properties_get_all_dict())

    def test_docstring(self) -> None:
        from pydoc import getdoc

        s = FreedesktopDbus(self.bus)

        with self.subTest('Method doc'):
            self.assertTrue(getdoc(s.get_connection_pid))

        with self.subTest('Property doc (through class dict)'):
            self.assertTrue(getdoc(s.__class__.__dict__['features']))

    def test_interface_composition(self) -> None:
        class OneInterface(
            DbusInterfaceCommon,
            interface_name="org.example.one",
        ):
            @dbus_method(result_signature="x")
            def one(self) -> int:
                raise NotImplementedError

        class TwoInterface(
            DbusInterfaceCommon,
            interface_name="org.example.two",
        ):
            @dbus_method(result_signature="t")
            def two(self) -> int:
                raise NotImplementedError

        class CombinedInterface(TwoInterface, OneInterface):
            ...

        test_combined = CombinedInterface("org.test", "/")
        test_combined_interfaces = [
            iface for iface, _ in test_combined._dbus_iter_interfaces_meta()
        ]

        # Verify the order of reported interfaces on the combined class.
        self.assertEqual(test_combined_interfaces, [
            "org.freedesktop.DBus.Peer",
            "org.freedesktop.DBus.Introspectable",
            "org.freedesktop.DBus.Properties",
            "org.example.one",
            "org.example.two",
        ])


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2023 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from asyncio import Event, get_running_loop
from asyncio import run as asyncio_run
from asyncio import sleep, wait_for
from asyncio.subprocess import create_subprocess_exec
from typing import TYPE_CHECKING
from unittest import SkipTest

from sdbus.exceptions import (
    DbusFailedError,
    DbusFileExistsError,
    DbusNoReplyError,
    DbusPropertyReadOnlyError,
    DbusUnknownObjectError,
    SdBusLibraryError,
    SdBusUnmappedMessageError,
)
from sdbus.sd_bus_internals import (
    DBUS_ERROR_TO_EXCEPTION,
    DbusPropertyEmitsChangeFlag,
)
from sdbus.unittest import IsolatedDbusTestCase
from sdbus.utils.parse import parse_properties_changed

from sdbus import (
    DbusInterfaceCommonAsync,
    DbusNoReplyFlag,
    dbus_method_async,
    dbus_method_async_override,
    dbus_property_async,
    dbus_property_async_override,
    dbus_signal_async,
    get_current_message,
)

if TYPE_CHECKING:
    from sdbus.dbus_proxy_async_interfaces import (
        DBUS_PROPERTIES_CHANGED_TYPING,
    )
else:
    DBUS_PROPERTIES_CHANGED_TYPING = None


class TestPing(IsolatedDbusTestCase):

    async def test_ping_with_busctl(self) -> None:
        try:
            busctl_process = await create_subprocess_exec(
                '/usr/bin/busctl',
                '--user',
                'call',
                'org.freedesktop.DBus', '/org/freedesktop/DBus',
                'org.freedesktop.DBus.Peer', 'Ping',
            )
        except FileNotFoundError:
            raise SkipTest('busctl not installed')

        return_code = await busctl_process.wait()
        self.assertEqual(return_code, 0)

    async def test_ping(self) -> None:
        m = self.bus.new_method_call_message(
            'org.freedesktop.DBus', '/org/freedesktop/DBus',
            'org.freedesktop.DBus.Peer', 'Ping',
        )
        r = await self.bus.call_async(m)
        self.assertIsNone(r.get_contents())


class TestRequestName(IsolatedDbusTestCase):
    async def test_request_name(self) -> None:
        await self.bus.request_name_async("org.example.test", 0)


TEST_INTERFACE_NAME = "org.test.test"


class TestInterface(
    DbusInterfaceCommonAsync,
    interface_name=TEST_INTERFACE_NAME,
):

    def __init__(self) -> None:
        super().__init__()
        self.test_string = 'test_property'
        self.test_string_read = 'read'
        self.test_no_reply_string = 'no'
        self.property_private = 100
        self.no_reply_sync = Event()

    @dbus_method_async("s", "s")
    async def upper(self, string: str) -> str:
        """Uppercase the input"""
        return string.upper()

    @dbus_method_async(result_signature='s')
    async def get_sender(self) -> str:
        message = get_current_message()
        return message.sender or ''

    @dbus_method_async(result_signature='x')
    async def test_int(self) -> int:
        return 1

    @dbus_method_async(result_signature='x', result_args_names=('an_int', ))
    async def int_annotated(self) -> int:
        return 1

    @dbus_property_async(
        "s",
        flags=DbusPropertyEmitsChangeFlag,
    )
    def test_property(self) -> str:
        """Test property"""
        return self.test_string

    @test_property.setter
    def test_property_set(self, new_property: str) -> None:
        self.test_string = new_property

    @dbus_property_async("s")
    def test_property_read_only(self) -> str:
        return self.test_string_read

    @dbus_property_async("x")
    def test_property_private(self) -> int:
        return self.property_private

    @test_property_private.setter_private
    def test_private_setter(self, new_value: int) -> None:
        self.property_private = new_value

    @dbus_method_async("sb", "s")
    async def kwargs_function(
            self,
            input: str = 'test',
            is_upper: bool = True) -> str:
        if is_upper:
            return input.upper()
        else:
            return input.lower()

    @dbus_method_async("sb", "s", 0, ('string_result', ))
    async def kwargs_function_annotated(
            self,
            input: str = 'test',
            is_upper: bool = True) -> str:
        if is_upper:
            return input.upper()
        else:
            return input.lower()

    @dbus_signal_async('ss')
    def test_signal(self) -> tuple[str, str]:
        """Test signal"""
        raise NotImplementedError

    @dbus_method_async()
    async def raise_base_exception(self) -> None:
        raise DbusFailedError('Test error')

    @dbus_method_async()
    async def raise_derived_exception(self) -> None:
        raise DbusFileExistsError('Test error 2')

    @dbus_method_async()
    async def raise_custom_error(self) -> None:
        raise DbusErrorTest('Custom')

    @dbus_method_async()
    async def raise_and_unmap_error(self) -> None:
        try:
            DBUS_ERROR_TO_EXCEPTION.pop('org.example.Nothing')
        except KeyError:
            ...

        raise DbusErrorUnmappedLater('Should be unmapped')

    @dbus_method_async()
    async def raise_python_exc(self) -> None:
        raise ValueError("Test!")

    @dbus_method_async('s', flags=DbusNoReplyFlag)
    async def no_reply_method(self, new_value: str) -> None:
        self.no_reply_sync.set()

    @dbus_property_async("s")
    def test_constant_property(self) -> str:
        return "a"

    @dbus_method_async(
        result_signature='(ss)'
    )
    async def test_struct_return(self) -> tuple[str, str]:
        return ('hello', 'world')

    @dbus_method_async(
        result_signature='(ss)'
    )
    async def test_struct_return_workaround(self) -> tuple[tuple[str, str]]:
        return (('hello', 'world'), )

    @dbus_method_async()
    async def looong_method(self) -> None:
        await sleep(100)

    @dbus_signal_async()
    def empty_signal(self) -> None:
        raise NotImplementedError

    @dbus_method_async()
    async def returns_none_method(self) -> None:
        return

    @dbus_method_async(
        input_signature="(iiii)",
        result_signature="i",
    )
    async def takes_struct_method(
        self,
        int_struct: tuple[int, int, int, int],
    ) -> int:
        a, b, c, d = int_struct
        return a*b*c*d

    @dbus_method_async("s", "x")
    async def return_length(self, input_str: str) -> int:
        return len(input_str)


class DbusErrorTest(DbusFailedError):
    dbus_error_name = 'org.example.Error'


class DbusErrorUnmappedLater(DbusFailedError):
    dbus_error_name = 'org.example.Nothing'


TEST_SERVICE_NAME = 'org.example.test'


def initialize_object() -> tuple[TestInterface, TestInterface]:
    test_object = TestInterface()
    test_object.export_to_dbus('/')

    test_object_connection = TestInterface.new_proxy(
        TEST_SERVICE_NAME, '/')

    return test_object, test_object_connection


class TestProxy(IsolatedDbusTestCase):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()
        await self.bus.request_name_async(TEST_SERVICE_NAME, 0)

    async def test_method_kwargs(self) -> None:
        test_object, test_object_connection = initialize_object()

        self.assertEqual(
            'TEST',
            await test_object_connection.kwargs_function(
                'test', True)
        )

        self.assertEqual(
            'TEST',
            await test_object_connection.kwargs_function())

        self.assertEqual(
            'test',
            await test_object_connection.kwargs_function(
                is_upper=False))

        self.assertEqual(
            'ASD',
            await test_object_connection.kwargs_function('asd'))

        self.assertEqual(
            'ASD',
            await test_object_connection.kwargs_function(input='asd'))

        self.assertEqual(
            'asd',
            await test_object_connection.kwargs_function(
                'ASD', is_upper=False))

        self.assertEqual(
            'asd',
            await test_object_connection.kwargs_function(
                input='ASD', is_upper=False))

    async def test_method(self) -> None:
        test_object, test_object_connection = initialize_object()

        test_string = 'asdafsrfgdrtuhrytuj'

        self.assertEqual(test_string.upper(),
                         await test_object.upper(test_string))

        self.assertEqual(1, await test_object_connection.test_int())

        self.assertEqual(
            test_string.upper(),
            await test_object_connection.upper(test_string))

        self.assertEqual(
            await wait_for(test_object.test_struct_return(), 0.5),
            await wait_for(test_object_connection.test_struct_return(), 0.5),
        )

        self.assertEqual(
            (await wait_for(
                test_object.test_struct_return_workaround(), 0.5))[0],
            await wait_for(
                test_object_connection.test_struct_return_workaround(), 0.5),
        )

        self.assertTrue(await test_object_connection.get_sender())

        with self.subTest("Test method that returns None"):
            self.assertIsNone(
                await test_object
                .returns_none_method()  # type: ignore[func-returns-value]
            )
            self.assertIsNone(
                await test_object_connection
                .returns_none_method()   # type: ignore[func-returns-value]
            )

        with self.subTest("Test method that takes a single struct"):
            self.assertEqual(
                await test_object.takes_struct_method((2, 3, 4, 5)),
                120,
            )
            self.assertEqual(
                await test_object_connection.takes_struct_method((9, 8, 7, 6)),
                3024,
            )

    async def test_subclass(self) -> None:
        test_object, test_object_connection = initialize_object()

        test_var = ['asdasd']

        class TestInheritence(TestInterface):
            @dbus_method_async_override()
            async def test_int(self) -> int:
                return 2

            @dbus_property_async_override()
            def test_property(self) -> str:
                return test_var[0]

            @test_property.setter
            def test_property_setter(self, var: str) -> None:
                test_var.insert(0, var)

        test_subclass = TestInheritence()

        test_subclass.export_to_dbus('/subclass', self.bus)

        self.assertEqual(await test_subclass.test_int(), 2)

        test_subclass_connection = TestInheritence.new_proxy(
            TEST_SERVICE_NAME, '/subclass', self.bus)

        self.assertEqual(await test_subclass_connection.test_int(), 2)

        self.assertEqual(test_var[0], await test_subclass.test_property)

        await test_subclass.test_property.set_async('12345')

        self.assertEqual(test_var[0], await test_subclass.test_property)
        self.assertEqual('12345', await test_subclass.test_property)

        with self.subTest('Test dbus to python mapping'):
            dbus_elements_map = (
                {
                    interface_name: meta.dbus_member_to_python_attr
                    for interface_name, meta in
                    TestInterface._dbus_iter_interfaces_meta()
                }
            )
            self.assertIn(
                "TestInt",
                dbus_elements_map[TEST_INTERFACE_NAME],
            )

            self.assertIn(
                "TestInt",
                dbus_elements_map[TEST_INTERFACE_NAME],
            )

            self.assertIn(
                "TestProperty",
                dbus_elements_map[TEST_INTERFACE_NAME],
            )

        with self.subTest('Tripple subclass'):
            class TestInheritenceTri(TestInheritence):
                @dbus_method_async_override()
                async def test_int(self) -> int:
                    return 3

                @dbus_property_async_override()
                def test_property(self) -> str:
                    return 'tri'

            test_subclass_tri = TestInheritenceTri()

            test_subclass_tri.export_to_dbus('/subclass/tri', self.bus)

            self.assertEqual(await test_subclass_tri.test_int(), 3)

            test_subclass_tri_connection = TestInheritenceTri.new_proxy(
                TEST_SERVICE_NAME, '/subclass/tri', self.bus)

            self.assertEqual(await test_subclass_tri_connection.test_int(), 3)

            self.assertEqual(await test_subclass_tri.test_property, 'tri')
            self.assertEqual(
                await test_subclass_tri_connection.test_property, 'tri')

    async def test_properties(self) -> None:
        test_object, test_object_connection = initialize_object()

        self.assertEqual(
            'test_property',
            await test_object.test_property.get_async())

        self.assertEqual(
            'test_property', await test_object.test_property)

        self.assertEqual(
            await wait_for(test_object_connection.test_property, 0.5),
            await test_object.test_property)

        self.assertEqual(
            'test_property',
            await wait_for(test_object_connection.test_property, 0.5))

        self.assertEqual(
            await test_object.test_property_read_only,
            await wait_for(
                test_object_connection.test_property_read_only, 0.5))

        new_string = 'asdsgrghdthdth'

        await wait_for(
            test_object_connection.test_property.set_async(
                new_string),
            0.5)

        self.assertEqual(
            new_string, await test_object.test_property)

        self.assertEqual(
            new_string,
            await wait_for(test_object_connection.test_property, 0.5)
        )

    async def test_signal(self) -> None:
        test_object, test_object_connection = initialize_object()

        test_tuple = ('sgfsretg', 'asd')

        async with self.assertDbusSignalEmits(
                test_object.test_signal
            ) as local_signals_record, self.assertDbusSignalEmits(
                test_object_connection.test_signal
        ) as remote_signals_record:
            test_object.test_signal.emit(test_tuple)

        async with self.assertDbusSignalEmits(
                test_object.test_signal
            ) as local_signals_record, self.assertDbusSignalEmits(
                test_object_connection.test_signal
        ) as remote_signals_record:
            test_object.test_signal.emit(test_tuple)

        self.assertEqual([test_tuple], local_signals_record.output)
        self.assertEqual([test_tuple], remote_signals_record.output)

    async def test_signal_catch_anywhere(self) -> None:
        test_object, test_object_connection = initialize_object()

        loop = get_running_loop()

        test_tuple = ('sgfsretg', 'asd')

        with self.subTest('Catch anywhere over D-Bus object'):
            async def catch_anywhere_oneshot_dbus(
            ) -> tuple[str, tuple[str, str]]:
                async for x in test_object_connection.test_signal\
                        .catch_anywhere():
                    return x

                raise RuntimeError

            catch_anywhere_over_dbus_task \
                = loop.create_task(catch_anywhere_oneshot_dbus())

            await sleep(0)

            test_object.test_signal.emit(test_tuple)

            self.assertEqual(
                ('/', test_tuple),
                await wait_for(catch_anywhere_over_dbus_task, timeout=1),
            )

        with self.subTest('Catch anywhere over D-Bus class'):
            async def catch_anywhere_oneshot_from_class(
            ) -> tuple[str, tuple[str, str]]:
                async for x in TestInterface.test_signal.catch_anywhere(
                        TEST_SERVICE_NAME, self.bus):
                    return x

                raise RuntimeError

            catch_anywhere_from_class_task \
                = loop.create_task(catch_anywhere_oneshot_from_class())

            await sleep(0)

            test_object.test_signal.emit(test_tuple)

            self.assertEqual(
                ('/', test_tuple),
                await wait_for(catch_anywhere_from_class_task, timeout=1),
            )

        with self.subTest('Catch anywhere over local object'):
            async def catch_anywhere_oneshot_local(
            ) -> tuple[str, tuple[str, str]]:
                async for x in test_object.test_signal.catch_anywhere():
                    return x

                raise RuntimeError

            catch_anywhere_over_local_task \
                = loop.create_task(catch_anywhere_oneshot_local())

            with self.assertRaises(NotImplementedError):
                await wait_for(
                    catch_anywhere_over_local_task,
                    timeout=1,
                )

    async def test_signal_multiple_readers(self) -> None:
        test_object, test_object_connection = initialize_object()

        loop = get_running_loop()

        test_tuple = ('sgfsretg', 'asd')

        async def reader_one() -> tuple[str, str]:
            async for x in test_object_connection.test_signal.catch():
                return test_tuple

            raise RuntimeError

        async def reader_two() -> tuple[str, str]:
            async for x in test_object_connection.test_signal.catch():
                return test_tuple

            raise RuntimeError

        t1 = loop.create_task(reader_one())
        t2 = loop.create_task(reader_two())

        loop.call_at(0, test_object.test_signal.emit, test_tuple)

        self.assertEqual(test_tuple, await wait_for(t1, timeout=1))
        self.assertEqual(test_tuple, await wait_for(t2, timeout=1))

    async def test_exceptions(self) -> None:
        test_object, test_object_connection = initialize_object()

        async def first_test() -> None:
            await test_object_connection.raise_base_exception()

        with self.assertRaises(DbusFailedError):
            await wait_for(first_test(), timeout=1)

        async def second_test() -> None:
            await test_object_connection.raise_derived_exception()

        with self.assertRaises(DbusFileExistsError):
            await wait_for(second_test(), timeout=1)

        async def third_test() -> None:
            await test_object_connection.raise_custom_error()

        with self.assertRaises(DbusErrorTest):
            await wait_for(third_test(), timeout=1)

        def bad_exception_error_name_used() -> None:
            class BadDbusError(DbusFailedError):
                dbus_error_name = 'org.freedesktop.DBus.Error.NoMemory'

        self.assertRaises(ValueError, bad_exception_error_name_used)

        def bad_exception_no_error_name() -> None:
            class BadDbusError(DbusFailedError):
                ...

        self.assertRaises(TypeError, bad_exception_no_error_name)

        async def test_unmapped_expection() -> None:
            await test_object_connection.raise_and_unmap_error()

        with self.assertRaises(SdBusUnmappedMessageError):
            await wait_for(test_unmapped_expection(), timeout=1)

    async def test_no_reply_method(self) -> None:
        test_object, test_object_connection = initialize_object()

        await wait_for(test_object_connection.no_reply_method('yes'),
                       timeout=1)

        await wait_for(test_object.no_reply_sync.wait(), timeout=1)

    async def test_interface_remove(self) -> None:
        test_object, test_object_connection = initialize_object()

        from gc import collect

        del test_object

        collect()

        with self.assertRaises(DbusUnknownObjectError):
            await wait_for(test_object_connection.dbus_introspect(),
                           timeout=0.2)

    def test_docstring(self) -> None:
        test_object, test_object_connection = initialize_object()

        from pydoc import getdoc

        self.assertTrue(getdoc(test_object.upper))

        self.assertTrue(getdoc(test_object_connection.upper))

        self.assertTrue(getdoc(test_object.test_property))

        self.assertTrue(
            getdoc(test_object_connection.test_property))

        self.assertTrue(getdoc(test_object.test_signal))

        self.assertTrue(
            getdoc(test_object_connection.test_signal))

    async def test_emits_properties_changed(self) -> None:
        test_object, test_object_connection = initialize_object()

        test_str = 'should_be_emited'

        loop = get_running_loop()

        async def catch_property_emit_connection() -> str:
            async for x in test_object_connection.properties_changed.catch():
                for v in x[1].values():
                    probably_str = v[1]
                    if isinstance(probably_str, str):
                        return probably_str
                    else:
                        raise TypeError
            raise ValueError

        async def catch_property_emit_local() -> str:
            async for x in test_object.properties_changed.catch():
                for v in x[1].values():
                    probably_str = v[1]
                    if isinstance(probably_str, str):
                        return probably_str
                    else:
                        raise TypeError
            raise ValueError

        t1 = loop.create_task(catch_property_emit_connection())
        t2 = loop.create_task(catch_property_emit_local())

        await test_object_connection.test_property.set_async(test_str)

        t1_result = await wait_for(t1, timeout=0.2)
        t2_result = await wait_for(t2, timeout=0.2)

        self.assertEqual(t1_result, test_str)
        self.assertEqual(t2_result, test_str)

    async def test_bus_close(self) -> None:
        test_object, test_object_connection = initialize_object()

        async def too_long_wait() -> None:
            await test_object_connection.looong_method()

        self.bus.close()

        with self.assertRaises(SdBusLibraryError):
            await wait_for(too_long_wait(), timeout=1)

    async def test_bus_timerfd(self) -> None:
        test_object, test_object_connection = initialize_object()

        self.bus.method_call_timeout_usec = 10_000  # 0.01 seconds

        loop = get_running_loop()

        start = loop.time()

        with self.assertRaises(DbusNoReplyError):
            await wait_for(test_object_connection.looong_method(), timeout=1)

        self.assertAlmostEqual(loop.time() - start, 0.01, delta=0.01)

    async def test_signal_queue_wildcard_match(self) -> None:
        test_object, test_object_connection = initialize_object()

        loop = get_running_loop()
        future = loop.create_future()

        slot = await self.bus.match_signal_async(
            TEST_SERVICE_NAME,
            None, None, None,
            future.set_result)

        try:
            test_object.test_signal.emit(('test', 'signal'))

            await wait_for(future, timeout=1)
            message = future.result()
            self.assertEqual(message.member, "TestSignal")
        finally:
            slot.close()

    async def test_class_with_string_subclass_parameter(self) -> None:
        from enum import Enum

        class InterfaceNameEnum(str, Enum):
            FOO = 'org.example.foo'
            BAR = 'org.example.bar'

        class ObjectPathEnum(str, Enum):
            FOO = '/foo'
            BAR = '/bar'

        class EnumedInterfaceAsync(
            DbusInterfaceCommonAsync,
            interface_name=InterfaceNameEnum.BAR,
        ):

            @dbus_property_async('s')
            def hello_world(self) -> str:
                return 'Hello World!'

        test_object = EnumedInterfaceAsync()
        test_object.export_to_dbus(ObjectPathEnum.FOO)

    async def test_properties_get_all_dict(self) -> None:
        test_object, test_object_connection = initialize_object()

        dbus_dict = await test_object_connection._properties_get_all(
            TEST_INTERFACE_NAME)

        self.assertEqual(
            await test_object.test_property,
            dbus_dict['TestProperty'][1],
        )

        self.assertEqual(
            await test_object.test_property,
            (
                await test_object_connection.properties_get_all_dict()
            )['test_property'],
        )

    async def test_empty_signal(self) -> None:
        test_object, test_object_connection = initialize_object()

        async with self.assertDbusSignalEmits(
                test_object.empty_signal
            ) as local_signals_record, self.assertDbusSignalEmits(
                test_object_connection.empty_signal
        ) as remote_signals_record:
            test_object.empty_signal.emit(None)

        self.assertEqual([None], local_signals_record.output)
        self.assertEqual([None], remote_signals_record.output)

    async def test_properties_changed(self) -> None:
        test_object, test_object_connection = initialize_object()

        test_str = 'should_be_emited'

        async with self.assertDbusSignalEmits(
            test_object_connection.properties_changed
        ) as properties_changed_catch:
            await test_object_connection.test_property.set_async(test_str)

        properties_changed_data = properties_changed_catch.output[0]

        parsed_dict_from_class = parse_properties_changed(
            TestInterface, properties_changed_data)
        self.assertEqual(
            test_str,
            parsed_dict_from_class['test_property'],
        )

        parsed_dict_from_object = parse_properties_changed(
            test_object_connection, properties_changed_data)
        self.assertEqual(
            test_str,
            parsed_dict_from_object['test_property'],
        )

        properties_changed_data[2].append('invalidated_property')
        parsed_dict_with_invalidation = parse_properties_changed(
            test_object, properties_changed_data,
            on_unknown_member='reuse',
        )
        self.assertIsNone(
            parsed_dict_with_invalidation['invalidated_property'])

    async def test_property_private_setter(self) -> None:
        test_object, test_object_connection = initialize_object()

        new_value = 200
        self.assertNotEqual(
            await test_object_connection.test_property_private,
            new_value
        )

        with self.assertRaises(DbusPropertyReadOnlyError):
            await test_object_connection.test_property_private.set_async(
                new_value)

        async with self.assertDbusSignalEmits(
            test_object_connection.properties_changed
        ) as properties_changed_catch:
            await test_object.test_property_private.set_async(new_value)

        changed_properties = properties_changed_catch.output[0]

        self.assertEqual(
            await test_object_connection.test_property_private,
            new_value
        )

        self.assertIn('TestPropertyPrivate', changed_properties[1])

    async def test_property_override_setter_private(self) -> None:

        test_int = 1

        class TestInterfacePrivateSetter(TestInterface):
            @dbus_property_async_override()
            def test_property_private(self) -> int:
                return test_int

            @test_property_private.setter_private
            def _private_setter(self, new_value: int) -> None:
                nonlocal test_int
                test_int = new_value

        test_object = TestInterfacePrivateSetter()
        test_object.export_to_dbus('/')
        test_object_connection = TestInterface.new_proxy(
            TEST_SERVICE_NAME, '/')

        self.assertEqual(
            await test_object_connection.test_property_private,
            test_int,
        )

        async def catch_properties_changed() -> int:
            async for x in test_object_connection.properties_changed:
                changed_attr = parse_properties_changed(
                    TestInterface, x)["test_property_private"]

                if not isinstance(changed_attr, int):
                    raise TypeError

                return changed_attr

            raise RuntimeError

        catch_changed_task = get_running_loop(
        ).create_task(catch_properties_changed())

        with self.assertRaises(DbusPropertyReadOnlyError):
            await test_object_connection.test_property_private.set_async(10)

        await test_object.test_property_private.set_async(10)

        self.assertEqual(
            await test_object_connection.test_property_private,
            10,
        )

        self.assertEqual(
            await test_object.test_property_private,
            test_int,
        )

        self.assertEqual(
            await wait_for(catch_changed_task, timeout=1),
            10,
        )

    async def test_interface_composition(self) -> None:
        class OneInterface(
            DbusInterfaceCommonAsync,
            interface_name="org.example.one",
        ):
            @dbus_method_async(result_signature="x")
            async def one(self) -> int:
                raise NotImplementedError

        class TwoInterface(
            DbusInterfaceCommonAsync,
            interface_name="org.example.two",
        ):
            @dbus_method_async(result_signature="t")
            async def two(self) -> int:
                return 2

        class CombinedInterface(TwoInterface, OneInterface):
            ...

        test_combined = CombinedInterface()
        test_combined_interfaces = [
            iface for iface, _ in test_combined._dbus_iter_interfaces_meta()
        ]

        # Verify the order of reported interfaces on the combined class.
        self.assertEqual(test_combined_interfaces, [
            "org.freedesktop.DBus.Peer",
            "org.freedesktop.DBus.Introspectable",
            "org.freedesktop.DBus.Properties",
            "org.example.one",
            "org.example.two",
        ])

    async def test_extremely_large_string(self) -> None:
        test_object, test_object_connection = initialize_object()

        extremely_large_string = "a" * 8423681

        remote_len = await wait_for(
            test_object_connection.return_length(
                extremely_large_string
            ),
            timeout=10,
        )

        self.assertEqual(
            remote_len,
            len(extremely_large_string),
        )

        # Check that calling regular methods still works.
        for _ in range(5):
            await test_object_connection.returns_none_method()

    async def test_export_handle(self) -> None:
        test_object = TestInterface()
        test_object_connection = TestInterface.new_proxy(
            TEST_SERVICE_NAME, '/',
        )
        with self.assertRaises(DbusUnknownObjectError):
            await test_object_connection.returns_none_method()

        with test_object.export_to_dbus("/"):
            await test_object_connection.returns_none_method()

        with self.assertRaises(DbusUnknownObjectError):
            await test_object_connection.returns_none_method()

        test_object2 = TestInterface()
        handle = test_object2.export_to_dbus("/")
        await test_object_connection.returns_none_method()
        handle.stop()

        with self.assertRaises(DbusUnknownObjectError):
            await test_object_connection.returns_none_method()

    def test_asyncio_run_different_loops(self) -> None:
        bus = self.bus

        async def test() -> None:
            dbus_object = DbusInterfaceCommonAsync.new_proxy(
                "org.freedesktop.DBus",
                "/org/freedesktop/DBus",
                bus,
            )
            await wait_for(dbus_object.dbus_ping(), timeout=1)

        with self.assertRaisesRegex(RuntimeError, "different loop"):
            asyncio_run(test())

    async def test_python_exc(self) -> None:
        test_object, test_object_connection = initialize_object()

        with self.assertRaisesRegex(ValueError, "Test!"):
            await test_object.raise_python_exc()

        with self.assertRaisesRegex(ValueError, "Test!"):
            await test_object_connection.raise_python_exc()

    async def test_empty_dbus_interface(self) -> None:
        class Empty(
            DbusInterfaceCommonAsync,
            interface_name="org.empty",
        ):
            ...

        empty_local = Empty()
        empty_local.export_to_dbus("/")
        empty_proxy = Empty.new_proxy(TEST_SERVICE_NAME, "/")

        intro = await empty_proxy.dbus_introspect()
        self.assertIn('<interface name="org.empty">', intro)



# SPDX-License-Identifier: LGPL-2.1-or-later

# Copyright (C) 2020-2022 igo95862

# This file is part of python-sdbus

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
from __future__ import annotations

from asyncio import get_running_loop, wait_for
from typing import Any

from sdbus.exceptions import DbusFailedError
from sdbus.unittest import IsolatedDbusTestCase

from sdbus import (
    DbusInterfaceCommonAsync,
    dbus_method_async,
    request_default_bus_name_async,
)

HELLO_WORLD = 'Hello, world!'


class DbusDeriveMethodError(DbusFailedError):
    dbus_error_name = 'org.example.Method.Error'


class IndependentError(Exception):
    ...


GOOD_STR = 'Good'


class InterfaceWithErrors(
    DbusInterfaceCommonAsync,
    interface_name='org.example.test',
):

    @dbus_method_async(result_signature='s')
    async def hello_independent_error(self) -> str:
        raise IndependentError

    @dbus_method_async(result_signature='s')
    async def hello_derrived_error(self) -> str:
        raise DbusDeriveMethodError


class TestHighLevelErrors(IsolatedDbusTestCase):
    async def asyncSetUp(self) -> None:
        await super().asyncSetUp()

        await request_default_bus_name_async('org.test')
        self.test_object = InterfaceWithErrors()
        self.test_object.export_to_dbus('/')

        self.test_object_connection = InterfaceWithErrors.new_proxy(
            'org.test', '/')

        loop = get_running_loop()

        def silence_exceptions(*args: Any, **kwrags: Any) -> None:
            ...

        loop.set_exception_handler(silence_exceptions)

    async def test_method_indenendent_error(self) -> None:
        with self.assertRaises(DbusFailedError) as cm:
            await wait_for(
                self.test_object_connection.hello_independent_error(),
                timeout=1,
            )

        exception = cm.exception
        self.assertIs(exception.__class__, DbusFailedError)

    async def test_method_derived_error(self) -> None:
        with self.assertRaises(DbusDeriveMethodError):
            await wait_for(
                self.test_object_connection.hello_derrived_error(),
                timeout=1,
            )
